#!/usr/bin/env/python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import abc
from typing import Tuple


class RendezvousClosedException(Exception):
    """
    Raised when a rendezvous for the specified run_id is closed.
    This is used to signal completion to nodes that arrive late.
    """

    pass


class RendezvousTimeoutException(Exception):
    """
    Raised from `next_rendezvous` to signal that the rendezvous did not
    succeed within the allocated time. This is meant to be interpreted
    as a non-retryable type of failure.
    """

    pass


class RendezvousNonRetryableError(Exception):
    """
    Raised from any of the `RendezvousHandler` methods when a failure
    occured that should not be retried with the same worker process.
    """

    pass


class RendezvousHandler(abc.ABC):
    """
    Main rendezvous interface.

    .. note:: torchelastic users normally **do not** need to implement their
              own ``RendezvousHandler``. An implementation based on
              `etcd <https://etcd.io/>`__ is already provided, and is recommended
              for most users, provided they can deploy it in their environment.

    .. warning:: torchelastic is currently considered experimental,
                 so the APIs may change!
    """

    @abc.abstractmethod
    # pyre-fixme[11]: Annotation `Store` is not defined as a type.
    def next_rendezvous(self) -> Tuple["torch.distributed.Store", int, int]:
        """
        Main entry-point into the rendezvous barrier.
        Blocks until the rendezvous is complete (and the current
        process is included in the formed worker group), or a timeout occurs, or
        rendezvous was marked closed.

        Returns a tuple of (``c10d Store``, ``rank``, ``world size``)

        Raises:
            ``RendezvousClosedException`` if rendezvous for the current
               job is closed.
            ``RendezvousTimeoutException`` on timeout
        """
        pass

    @abc.abstractmethod
    def is_closed(self) -> bool:
        """
        Checks whether rendezvous for current job has been closed,
        which means all future attempts to re-rendezvous (within same job) will
        fail.

        .. note:: ``is_closed``/``set_closed`` have semantics of eventual
                  propagation, and should not be used for synchronization.
                  The intention here is that if at least one worker decides
                  the job is finished, it will close the rendezvous, and
                  other workers will soon observe this and stop
                  training/rendezvous-ing as well.
        """
        pass

    @abc.abstractmethod
    def set_closed(self):
        """
        Used to mark the rendezvous (for current job) as closed.
        """
        pass

    @abc.abstractmethod
    def num_nodes_waiting(self) -> int:
        """
        Returns number of workers who *arrived late* at
        the rendezvous barrier, hence weren’t included in the current worker
        group.

        Callers should periodically call this method to check whether
        new members are waiting to join the job and if so admit them by
        calling ``next_rendezvous()`` (re-rendezvous).
        """
        pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import abc
import logging
import socket
import time
from contextlib import closing
from enum import Enum
from typing import Any, Callable, Dict, Tuple

import torchelastic.rendezvous as rdzv
from torchelastic.metrics import prof, put_metric


DEFAULT_ROLE = "default"

log = logging.getLogger(__name__)


class WorkerSpec:
    """
    Contains blueprint information about a particular type of worker.
    For a given role, there must only exist a single worker spec.
    Worker spec is expected to be homogenous across all nodes (machine),
    that is each node runs the same number of workers for a particular spec.
    """

    __slots__ = [
        "role",
        "local_world_size",
        "fn",
        "args",
        "rdzv_handler",
        "max_restarts",
        "monitor_interval",
        "master_port",
    ]

    def __init__(
        self,
        role: str,
        local_world_size: int,
        fn: Callable,
        args: Tuple,
        rdzv_handler: rdzv.RendezvousHandler,
        max_restarts: int = 100,
        monitor_interval: float = 5.0,
        master_port=None,
    ):
        r"""

        Arguments:
            role (str): user-defined role for the workers with this spec
            local_world_size (int): number local workers to run
            fn (Callable): worker main entry point function
            args (Tuple): arguments to pass to ``fn(args)``
            rdzv_handler (RendezvousHandler): handles rdzv for this set of workers
            max_restarts (int): number of max retries for the workers
            monitor_interval (int): monitor status of workers every ``n`` seconds
            master_port (int): fixed port to run the c10d store on rank 0
                               if not specified then will chose a random free port
        """

        assert local_world_size > 0
        assert max_restarts > 0
        assert monitor_interval > 0

        # Note: role is not used for data parallel, every worker has the same role
        # wiring it in to handle more elaborate situations later
        self.role = role
        self.local_world_size = local_world_size
        self.fn = fn
        self.args = args
        self.rdzv_handler = rdzv_handler
        self.max_restarts = max_restarts
        self.monitor_interval = monitor_interval
        self.master_port = master_port


class Worker:
    """
    Represents a worker instance. Contrast this with ``WorkerSpec`` that
    represents the specifications of a worker. A ``Worker`` is created from
    a ``WorkerSpec``. A ``Worker`` is to a ``WorkerSpec`` as an object is to
    a class.

    The ``id`` of the worker is interpreted
    by the specific implementation of ``ElasticAgent``. For a local
    agent, it could be the ``pid (int)`` of the worker, for a remote
    agent it could be encoded as ``host:port (string)``.

    Arguments:
        id (Any): uniquely identifies a worker (interpreted by the agent)
        local_rank (int): local rank of the worker
        global_rank (int): global rank of the worker
        world_size (int): number of workers (globally)
    """

    __slots__ = ["id", "local_rank", "global_rank", "world_size"]

    def __init__(self, local_rank: int):
        # unique identifier for this worker
        self.id: Any = None

        # rank of the worker among workers with the same role being monitored
        # by the same ``agent`` instance.
        self.local_rank: int = local_rank

        #  rank of the worker among all the workers with the same role
        #  across all ``agent`` instances.
        #  Global rank is not stable between re-rendezvous.
        # pyre-fixme[8]: Attribute has type `int`; used as `None`.
        self.global_rank: int = None

        # total number of workers (globally). Due to elasticity
        # the world size may change between re-rendezvous.
        # pyre-fixme[8]: Attribute has type `int`; used as `None`.
        self.world_size: int = None


class WorkerState(Enum):
    """
    State of the ``WorkerGroup``. Workers in a worker groupchange state as a unit.
    If a single worker in a worker groupfails the entire set is considered
    failed::

      UNKNOWN - agent lost track of worker group state, unrecoverable
      INIT - worker group object created not yet started
      HEALTHY - workers running and healthy
      UNHEALTHY - workers running and unhealthy
      STOPPED - workers stopped (interruped) by the agent
      SUCCEEDED - workers finished running (exit 0)
      FAILED - workers failed to successfully finish (exit !0)


    A worker group starts from an initial ``INIT`` state,
    then progresses to ``HEALTHY`` or ``UNHEALTHY`` states,
    and finally reaches a terminal ``SUCCEEDED`` or ``FAILED`` state.

    Worker groups can be interrupted and temporarily put into ``STOPPED`` state
    by the agent. Workers in ``STOPPED`` state are scheduled to be restarted
    in the near future by the agent. Some examples of workers being put into
    ``STOPPED`` state are:

    1. Worker group failure|unhealthy observed
    2. Membership change detected

    When actions (start, stop, rdzv, retry, etc) on worker group fails
    and results in the action being partially applied to the worker group
    the state will be ``UNKNOWN``. Typically this happens on uncaught/unhandled
    exceptions during state change events on the agent. The agent is not
    expected to recover worker groups in ``UNKNOWN`` state and is better off
    self terminating and allowing the job manager to retry the node.
    """

    UNKNOWN = 0
    INIT = 1
    HEALTHY = 2
    UNHEALTHY = 4
    STOPPED = 8
    SUCCEEDED = 16
    FAILED = 32

    @staticmethod
    def is_running(state: "WorkerState") -> bool:
        """
        Returns:
             `` True`` if the worker state represents workers still running
              (e.g. that the process exists but not necessarily healthy).
        """
        return state in {WorkerState.HEALTHY, WorkerState.UNHEALTHY}


class WorkerGroup:
    """
    Represents the set of ``Worker`` instances for the given ``WorkerSpec``
    managed by ``ElasticAgent``. Whether the worker groupcontains cross
    instance workers or not depends on the implementation of the agent.
    """

    __slots__ = ["spec", "workers", "store", "group_rank", "group_world_size", "state"]

    def __init__(self, spec: WorkerSpec):
        self.spec = spec
        self.workers = [Worker(local_rank=i) for i in range(self.spec.local_world_size)]

        # assigned after rdzv
        self.store = None
        self.group_rank = None
        self.group_world_size = None

        self.state = WorkerState.INIT


class MonitorResult:
    """
    Returned by the agent's ``_monitor_workers`` API. A holder object
    that holds information about the monitoring results.
    The ``ret_vals`` and ``exceptions`` field map each worker's
    return value (output) and exceptions (if any) accordingly by
    the workers global rank.

    ``state = SUCCEEDED`` will have ``ret_val``.
    ``state = FAILED`` will have ``exceptions``.
    For other states both these fields will be empty.
    """

    __slots__ = ["state", "ret_vals", "exceptions"]

    def __init__(
        self,
        state: WorkerState,
        # pyre-fixme[9]: ret_vals has type `Dict[int, typing.Any]`; used as `None`.
        ret_vals: Dict[int, Any] = None,
        # pyre-fixme[9]: exceptions has type `Dict[int, Exception]`; used as `None`.
        exceptions: Dict[int, Exception] = None,
    ):
        self.state = state
        # pyre-fixme[8]: Attribute has type `Dict[int, typing.Any]`; used as `None`.
        self.ret_vals = ret_vals
        self.exceptions = exceptions


class WorkerGroupFailureException(Exception):
    """
    Thrown when the agent cannot or has given up trying to run the workers.
    This is typically thrown:

    1. Exceeded ``max_restarts``.
    2. Workers fail with errors that are deemed ``NonRestartable``

    When constructing this exception the underlying worker exceptions
    are provided as a map of the worker's global rank to the exception.
    """

    def __init__(self, msg: str, worker_excs: Dict[int, Exception]):
        super().__init__(msg)
        self._worker_excs = worker_excs

    def get_worker_exceptions(self) -> Dict[int, Exception]:
        return self._worker_excs


def _get_socket_with_port() -> socket.socket:
    """
    Returns a free port on localhost that is "reserved" by binding a temporary
    socket on it. Close the socket before passing the port to the entity
    that requires it. Usage example

    ::

    sock = _get_socket_with_port()
    with closing(sock):
        port = sock.getsockname()[1]
        sock.close()
        # there is still a race-condition that some other process
        # may grab this port before func() runs
        func(port)
    """

    # pyre-fixme[28]: Unexpected keyword argument `type`.
    addrs = socket.getaddrinfo(
        host="localhost", port=None, family=socket.AF_UNSPEC, type=socket.SOCK_STREAM
    )
    for addr in addrs:
        family, type, proto, _, _ = addr
        s = socket.socket(family, type, proto)
        try:
            s.bind(("localhost", 0))
            s.listen(0)
            return s
        except OSError as e:
            s.close()
            log.info("Socket creation attempt failed.", exc_info=e)
    raise RuntimeError("Failed to create a socket")


def _get_fq_hostname() -> str:
    return socket.getfqdn(socket.gethostname())


class ElasticAgent(abc.ABC):
    """
    Agent process responsible for managing one or more worker processes.
    The worker processes are assumed to be regular distributed PyTorch scripts.
    When the worker process is created by the agent, the agent provides the
    necessary information for the worker processes to properly initialize
    a torch process group.

    The exact deployment topology and ratio of of agent-to-worker is dependent
    on the specific implementation of the agent and the user's job placement
    preferences. For instance, to run a distributed training job on GPU with
    8 trainers (one per GPU) one can:

    1. Use 8 x single GPU instances, place an agent per instance, managing
       1 worker per agent.
    2. Use 4 x double GPU instances, place an agent per instance, managing
       2 workers per agent.
    3. Use 2 x quad GPU instances, place an agent per instance, managing
       4 workers per agent.
    4. Use 1 x 8 GPU instance, place an agent per instance, managing
       8 workers per agent.

    Usage
    ::

     try:
         results = agent.run()
         return results[0] # return rank 0's results
     except WorkerGroupFailureException as e:
         exceptions = e.get_worker_exceptions()
         log.exception(f"worker 0 failed with: {exceptions[0]}")
     except Exception as e:
         log.exception(f"error while running agent")

    """

    @abc.abstractmethod
    def run(self, role: str = DEFAULT_ROLE) -> Dict[int, Any]:
        """
        Runs the agent, retrying the worker group on failures up to
        ``max_restarts``.

        Returns:
            The return values for each worker mapped by the worker's global rank.
            Empty if workers have void signature.

        Raises:
            WorkerGroupFailureException - workers did not successfully run
            Exception - any other failures NOT related to worker process
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_worker_group(self, role: str = DEFAULT_ROLE) -> WorkerGroup:
        """
        Returns:
            The ``WorkerGroup`` for the given ``role``.
            Note that the worker group is a mutable object and hence in a
            multi-threaded/process environment it may change state.
            Implementors are encouraged (but not required) to return
            a defensive read-only copy.
        """
        raise NotImplementedError()


class SimpleElasticAgent(ElasticAgent):
    """
    An ``ElasticAgent`` that manages workers (``WorkerGroup``)
    for a single ``WorkerSpec`` (e.g. one particular type of worker role).
    """

    def __init__(self, spec: WorkerSpec):
        self._worker_group = WorkerGroup(spec)
        self._remaining_restarts = self._worker_group.spec.max_restarts

    # pyre-fixme[14]: `get_worker_group` overrides method defined in `ElasticAgent`
    #  inconsistently.
    def get_worker_group(self) -> WorkerGroup:
        # TODO return an RO copy (need to create an ROWorkerGroup and ROWorkerSpec
        # since both these classes contain non-pure-data pointers - e.g. rdzv_handler)
        return self._worker_group

    @abc.abstractmethod
    def _start_workers(self, worker_group: WorkerGroup) -> Dict[int, Any]:
        r"""
        Starts ``worker_group.spec.local_world_size`` number of workers
        according to worker spec for the worker group .

        Returns a map of ``local_rank`` to worker ``id``.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _stop_workers(self, worker_group: WorkerGroup) -> None:
        r"""
        Stops all workers in the given worker group. Implementors
        must deal with workers in all states defined by ``WorkerState``.
        That is, it must gracefully handle stopping non-existent workers,
        unhealthy (stuck) workers, etc.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _monitor_workers(self, worker_group: WorkerGroup) -> MonitorResult:
        r"""
        Checks on the workers for the ``worker_group`` and returns
        the new state of the worker group.
        """
        raise NotImplementedError()

    @staticmethod
    def _set_master_addr_port(store, master_port):
        if master_port is None:
            sock = _get_socket_with_port()
            with closing(sock):
                master_port = sock.getsockname()[1]

        store.set("MASTER_ADDR", _get_fq_hostname().encode(encoding="UTF-8"))
        store.set("MASTER_PORT", str(master_port).encode(encoding="UTF-8"))

    @staticmethod
    def _get_master_addr_port(store) -> Tuple[str, int]:
        master_addr = store.get("MASTER_ADDR").decode(encoding="UTF-8")
        master_port = int(store.get("MASTER_PORT").decode(encoding="UTF-8"))
        return (master_addr, master_port)

    @prof
    def _rendezvous(self, worker_group: WorkerGroup) -> None:
        r"""
        Runs rendezvous for the workers specified by worker spec.
        Assigns workers a new global rank and world size.
        Updates the rendezvous store for the worker group.
        """

        spec = worker_group.spec
        stride = spec.local_world_size

        store, group_rank, group_world_size = spec.rdzv_handler.next_rendezvous()
        world_size = group_world_size * spec.local_world_size

        worker_group.store = store
        worker_group.group_rank = group_rank
        worker_group.group_world_size = group_world_size

        if group_rank == 0:
            self._set_master_addr_port(store, spec.master_port)

        assigned_global_ranks = []
        for worker in worker_group.workers:
            global_rank = (group_rank * stride) + worker.local_rank
            worker.global_rank = global_rank
            worker.world_size = world_size
            assigned_global_ranks.append(global_rank)

        master_addr, master_port = self._get_master_addr_port(store)
        restart_count = spec.max_restarts - self._remaining_restarts
        log.info(
            f"[{spec.role}] Rendezvous complete for workers.\n"
            f"Result:\n"
            f"\trestart_count={restart_count}\n"
            f"\tgroup_rank={group_rank}\n"
            f"\tgroup_world_size={group_world_size}\n"
            f"\trank stride={stride}\n"
            f"\tassigned global_ranks={assigned_global_ranks}\n"
            f"\tmaster_addr={master_addr}\n"
            f"\tmaster_port={master_port}\n"
        )

    @prof
    def _initialize_workers(self, worker_group: WorkerGroup) -> None:
        r"""
        Starts a fresh set of workers for the worker_group.
        Essentially a rendezvous followed by a start_workers.

        The caller should first call ``_stop_workers()`` to stop running workers
        prior to calling this method.

        Optimistically sets the state of the worker group that
        just started as ``HEALTHY`` and delegates the actual monitoring
        of state to ``_monitor_workers()`` method
        """
        role = worker_group.spec.role
        log.info(f"[{role}] Rendezvous'ing worker group")

        # TODO after stopping workers, wait at least monitor_interval*2 for
        # workers on different nodes to fail on a collective op before waiting
        # on the rdzv barrier, this way we ensure that nodes enter rdzv
        # at around the same time and reduce false positive rdzv timeout errors
        self._rendezvous(worker_group)

        log.info(f"[{role}] Starting worker group")
        worker_ids = self._start_workers(worker_group)
        for local_rank, id in worker_ids.items():
            worker = worker_group.workers[local_rank]
            worker.id = id

        worker_group.state = WorkerState.HEALTHY

    @prof
    def _restart_workers(self, worker_group: WorkerGroup) -> None:
        """
        Restarts (stops, rendezvous, starts) all local workers in the group.
        """

        role = worker_group.spec.role
        log.info(f"[{role}] Stopping worker group")
        self._stop_workers(worker_group)
        worker_group.state = WorkerState.STOPPED
        self._initialize_workers(worker_group)

    def run(self, role: str = DEFAULT_ROLE) -> Dict[int, Any]:
        # NOTE: currently only works for a single role

        spec = self._worker_group.spec
        role = spec.role

        log.info(f"[{role}] starting workers for function: {spec.fn.__name__}")

        self._initialize_workers(self._worker_group)
        monitor_interval = spec.monitor_interval
        rdzv_handler = spec.rdzv_handler

        while True:
            assert self._worker_group.state != WorkerState.INIT
            time.sleep(monitor_interval)
            monitor_result = self._monitor_workers(self._worker_group)
            state = monitor_result.state
            self._worker_group.state = state

            put_metric(f"workers.{role}.remaining_restarts", self._remaining_restarts)
            put_metric(f"workers.{role}.{state.name.lower()}", 1)

            if state == WorkerState.SUCCEEDED:
                log.info(f"[{role}] All workers successfully finished.")
                return monitor_result.ret_vals
            elif state in {WorkerState.UNHEALTHY, WorkerState.FAILED}:
                if self._remaining_restarts > 0:
                    log.info(
                        f"[{role}] Worker group {state.name}. "
                        f"{self._remaining_restarts}/{spec.max_restarts} attempts left;"
                        f" will restart worker group"
                    )
                    self._remaining_restarts -= 1
                    self._restart_workers(self._worker_group)
                else:
                    self._stop_workers(self._worker_group)
                    self._worker_group.state = WorkerState.FAILED
                    raise WorkerGroupFailureException(
                        f"[{role}] exceeded max_restarts={spec.max_restarts}",
                        monitor_result.exceptions,
                    )
            elif state == WorkerState.HEALTHY:
                # membership changes do not count as retries
                num_nodes_waiting = rdzv_handler.num_nodes_waiting()
                group_rank = self._worker_group.group_rank
                if num_nodes_waiting > 0:
                    log.info(
                        f"[{role}] Detected {num_nodes_waiting} "
                        f"new nodes from group_rank={group_rank}; "
                        f"will restart worker group"
                    )
                    self._restart_workers(self._worker_group)
            else:
                raise Exception(f"[{role}] Worker group in {state.name} state")



#!/usr/bin/env python3

import logging
from typing import Any, Dict, List

import fblearner.flow.api as flow
import torchelastic.rendezvous  # noqa F401
import torchelastic.rendezvous.parameters as parameters
from fblearner.flow.core.future import fut
from fblearner.flow.core.operatorregistry import LexicalContext
from torchelastic import metrics
from torchelastic.agent.server.api import WorkerGroupFailureException, WorkerSpec
from torchelastic.agent.server.local_elastic_agent import LocalElasticAgent


logger = logging.getLogger(__name__)


class MissingElasticDecoratorException(Exception):
    """
    Exception that is raised the ElasticLaunch is invoked on the function
    that is not annotated with the elastic decorator.
    """

    pass


class LaunchConfig:
    __slots__ = [
        "min_nodes",
        "max_nodes",
        "nproc_per_node",
        "rdzv_backend",
        "rdzv_endpoint",
        "run_id",
        "role",
        "rdzv_timeout",
        "max_restarts",
        "monitor_interval",
        "function_start_method",
        "gang_schedule",
        "flow_context",
        "flow_context_params",
    ]

    def __init__(
        self,
        min_nodes: int,
        max_nodes: int,
        nproc_per_node: int,
        rdzv_endpoint: str = "",
        rdzv_backend: str = "zeus-adapter",
        # pyre-fixme[9]: run_id has type `str`; used as `None`.
        run_id: str = None,
        role: str = "default_role",
        rdzv_timeout=300,
        max_restarts=10,
        monitor_interval=60,
        function_start_method="spawn",
        gang_schedule=True,
        flow_context=None,
        flow_context_params=None,
    ):
        """
        min_nodes (int): Minimum amount of nodes that the user function will
                         be launched on. Elastic agent ensures that the user
                         function start only when the min_nodes amount enters
                         the rendezvous.
        max_nodes (int): Maximum amount of nodes that the user function
                         will be launched on.
        nproc_per_node (int): On each node the elastic agent will launch
                              this amount of workers that will execute user
                              defined function.
        rdzv_backend: rdzv_backend to use in the rendezvous (zeus-adapter, etcd).
        rdzv_endpoint: The endpoint of the rdzv sync. storage.
        rdzv_id: The unique run id of the job (defaults to flow workflow run id).
        role: User defined role of the worker (defaults to "trainer")
        max_restarts: The maximum amount of restarts that elastic agent will conduct
                      on workers before failure.
        monitor_interval: The interval in seconds that is used by the elastic_agent
                          as a period of monitoring workers.
        function_start_method: The method is used by the elastic agent
                               to start the workers (spawn, fork, forkserver).
        gang_schedule: If True, schedules the nodes as a gang, if False, uses
                       pseudo-gangs. Gangs have an all-or-nothing semantic,
                       all nodes should be available to start running the job,
                       if one node fails all nodes are torn down and the job
                       restarts with the restart policy defined in the workflow.
                       Pseudo-gangs behave similar to parallel operators,
                       the job starts running as soon as one node is available
                       (but waits on rendezvous for min_nodes). If a node fails,
                       only that node is restarted. Note that pseudo-gangs can
                       cause deadlocks in the entitlement if jobs are not using
                       elasticity (all jobs will hold nodes and wait for min_nodes
                       and eventually fail on rendezvous timeout).
        flow_context: Implementation of fblearner.flow.core.operatorregistry.
        flow_context_params: Parameters that will be used to instantiate
                             fblearner.flow.core.operatorregistry.
                             Note: flow_context and flow_context_params are
                             mutually exclusive parameters.

        """
        self.min_nodes = min_nodes
        self.max_nodes = max_nodes
        self.nproc_per_node = nproc_per_node
        self.rdzv_backend = rdzv_backend
        self.rdzv_endpoint = rdzv_endpoint
        if run_id is None:
            self.run_id = flow.get_flow_environ().workflow_run_id
        else:
            self.run_id = run_id
        self.role = role
        self.rdzv_timeout = rdzv_timeout
        self.max_restarts = max_restarts
        self.monitor_interval = monitor_interval
        self.function_start_method = function_start_method
        self.gang_schedule = gang_schedule
        self.flow_context = flow_context
        self.flow_context_params = flow_context_params

        assert 0 < self.min_nodes <= self.max_nodes
        assert self.nproc_per_node > 0

        if flow_context is not None and flow_context_params is not None:
            raise ValueError(
                "Set up flow_context OR flow_context_params. "
                "You cannot set up both, since flow_context "
                "will be set up from flow_context_params."
            )


def _merge_worker_return_values(*return_values):
    """
    workaround to deal with the fact that the return value of each
    agent operator is a Future. Takes return values (Future)
    (one for each agent) and returns a merged map of all return values
    for each worker. Note that the return value for the agent operator
    is a map of the local workers (mapped by global rank -> worker ret_val)
    See: https://fburl.com/qa/k1zrx6if
    """
    merged_ret_vals: Dict[int, Any] = {}
    for rv in return_values:
        merged_ret_vals.update(rv)
    return merged_ret_vals


class elastic_launch:
    """
    Launches an "elastic" gang where the gang size may vary between
    min and max. Each gang node runs as an ``elastic_operator``.
    The ``elastic_operator`` runs one or more local workers (user defined ``fn``).

    On each ``elastic_operator``, an ``ElasticAgent`` is created to manage
    the local workers. When workers fail, the agent restarts ALL local workers
    up to ``max_restarts`` as configured in ``LaunchConfig``. When all
    restarts have been exhausted or the agent itself fails then the node
    (flow operator) fails and respects the retry policy that is configured
    on the operator and workflow.

    NOTE:
        1. Pass the ``fn`` arguments as non ``kwargs`` (e.g. no named parameters)
        2. ``fn`` must NOT have any decorators
        3. Pass additional flow operator arguments as ``kwargs``
        4. ``elastic_operator`` has the operator param ``use_forkserver=True``
        5. The return value is a Future resolving to a map of each worker's
           output mapped by their respective global rank.

    Usage

    ::

    def worker_fn(foo):
        torch.distributed.init_process_group(backend="gloo")
        # ...

    @flow.registered()
    @flow.typed()
    def main():
        outputs = elastic_launch(LaunchConfig, worker_fn)(
            foo,
            resource_requirements=ResourceRequirements,
            parse_macros=True)

        # return rank 0's output
        return outputs[0]


    Invalid Usage

    ::

    @flow.flow_async() # INVALID; worker_fn should not be decorated
    def worker_fn(foo):
        pass

    @flow.registered()
    @flow.typed()
    def main():
        outputs = elastic_launch(LauncConfig, worker_fn)(
            foo = foo, # INVALID; must NOT pass worker_fn args as kwargs
        )

        outputs[0].bar # INVALID; cannot access Future's attributes
    """

    def __init__(self, config: LaunchConfig, fn):
        self._config = config
        self._fn = fn

    def __call__(self, *args, **kwargs):
        # holds return values of each gang node
        return_values = []
        with self._lexical_context():
            for i in range(self._config.max_nodes):
                logging.info(f"Launching gang node {i}")
                ret = elastic_operator(self._config, self._fn, list(args), **kwargs)
                return_values.append(ret)
        return fut(_merge_worker_return_values)(*return_values)

    def _lexical_context(self):
        if self._config.flow_context is not None:
            return self._config.flow_context
        fn_name = self._fn.__qualname__
        gang_schedule = self._config.gang_schedule
        gang_name = (
            f"elastic_gang:{fn_name}"
            if gang_schedule
            else f"elastic_pseudo_gang:{fn_name}"
        )
        lc = LexicalContext(gang_name, gang_schedule=gang_schedule)
        if self._config.flow_context_params is not None:
            for param_name, param_value in self._config.flow_context_params.items():
                if hasattr(lc, param_name):
                    setattr(lc, param_name, param_value)
        return lc


@flow.flow_async(saves_partial_results=True)
@flow.typed()
def elastic_operator(config: LaunchConfig, fn, fn_args: List[Any]) -> Dict[int, Any]:
    logger.info(
        f"Starting elastic_operator with launch configs:\n"
        f"\tuser fn: {fn.__name__}\n"
        f"\tmin_nodes: {config.min_nodes}\n"
        f"\tmax_nodes: {config.max_nodes}\n"
        f"\tnproc_per_node: {config.nproc_per_node}\n"
        f"\trun_id: {config.run_id}\n"
        f"\trdzv_backend: {config.rdzv_backend}\n"
        f"\trdzv_endpoint: {config.rdzv_endpoint}\n"
        f"\tmax_restarts: {config.max_restarts}\n"
        f"\tmonitor_interval: {config.monitor_interval}"
    )

    rdzv_parameters = parameters.RendezvousParameters(
        config.rdzv_backend,
        config.rdzv_endpoint,
        config.run_id,
        config.min_nodes,
        config.max_nodes,
        config=(
            f"timeout={config.rdzv_timeout}"
            f",min_nodes_timeout={config.rdzv_timeout}"
            ",waitfor_previous_rendezvous=False"
        ),
    )
    rdzv_handler = parameters.get_rendezvous(rdzv_parameters)

    spec = WorkerSpec(
        role=config.role,
        local_world_size=config.nproc_per_node,
        fn=fn,
        args=tuple(fn_args),
        rdzv_handler=rdzv_handler,
        max_restarts=config.max_restarts,
        monitor_interval=config.monitor_interval,
    )

    metrics.initialize_metrics()
    agent = LocalElasticAgent(spec, config.function_start_method)

    try:
        return agent.run()
    except WorkerGroupFailureException as e:
        # user-code error; propagate the one from the smallest
        # rank on this node
        excs = e.get_worker_exceptions()
        min_rank = min(excs.keys())
        raise excs[min_rank]
    except Exception as e:
        # this is an agent (platform) error
        # TODO clearly distingush and categorize user vs platform errors
        raise e



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import ctypes
import multiprocessing
import queue
import uuid


def is_asan():
    """Determines if the Python interpreter is running with ASAN"""
    return hasattr(ctypes.CDLL(""), "__asan_init")


def is_tsan():
    """Determines if the Python interpreter is running with TSAN"""
    return hasattr(ctypes.CDLL(""), "__tsan_init")


def is_asan_or_tsan():
    return is_asan() or is_tsan()


def _get_or_raise(qout, qerr):
    # Helper function that either returns the value on the out queue
    # or raises the exception on the error queue.
    while True:
        try:
            return qout.get(False, 0.001)
        except queue.Empty:
            pass
        try:
            raise qerr.get(False, 0.001)
        except queue.Empty:
            pass


class TestCommon:
    def setUp(self):
        self.processes = []

    def tearDown(self):
        for p in self.processes:
            p.terminate()

    def _generate_run_id(self):
        return str(uuid.uuid4().int)

    def _spawn(self, fn, *args):
        name = "process #%d" % len(self.processes)
        qin = multiprocessing.Queue()
        qout = multiprocessing.Queue()
        qerr = multiprocessing.Queue()
        qio = (qin, qout, qerr)
        args = qio + (fn,) + args
        process = multiprocessing.Process(target=self._run, name=name, args=args)
        process.start()
        self.processes.append(process)
        return qio

    def _wait_all_and_clean(self):
        for p in self.processes:
            p.join()
        self.processes = []

    def _run(self, qin, qout, qerr, fn, *args):
        try:
            qout.put(fn(qin, *args))
        except Exception as e:
            qerr.put(e)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
from enum import Enum, unique

from jinja2 import Template
from util import wait_for


log = logging.getLogger(__name__)


@unique
class Accelerator(Enum):
    NONE = 0
    GPU = 1

    @classmethod
    def get_accelerator(cls, instance_type):
        """
        get_accelerator("p3.2xlarge") returns Accelerator.GPU
        get_accelerator("i3.xlarge") returns Accelerator.NONE
        """

        instance_accelerators = {
            "g2": Accelerator.GPU,
            "g3": Accelerator.GPU,
            "g4": Accelerator.GPU,
            "p2": Accelerator.GPU,
            "p3": Accelerator.GPU,
        }

        instance_family = instance_type[0:2]
        return instance_accelerators.get(instance_family, Accelerator.NONE)

    @classmethod
    def from_str(cls, accelerator_str):
        """
        returns the enum Accelerator value from a string representation
        """
        accelerators = {"none": Accelerator.NONE, "gpu": Accelerator.GPU}
        return accelerators.get(accelerator_str.lower(), Accelerator.NONE)

    def describe(self):
        """
        Returns a string representation of the enum.
        This method is intended to be used to label certain AWS
        resources in their descriptions/names for informative purposes

        e.g. launch template created for GPUs can be named as: torchelastic_gpu
        """

        string_rep = {Accelerator.NONE.value(): "cpu", Accelerator.GPU.value(): "gpu"}
        return string_rep.get(self, "unknown_accelerator")


class AutoScalingGroup:
    def __init__(self, session):
        self._session = session
        self._asg = session.client("autoscaling")
        self._ec2 = session.client("ec2")

    def get_user_data(self, user_data_template, **kwargs):
        if os.path.isabs(user_data_template):
            user_data_path = user_data_template
        else:
            user_data_path = os.path.join(os.path.dirname(__file__), user_data_template)

        with open(user_data_path) as f:
            user_data_template = Template(f.read())
            user_data = user_data_template.render(**kwargs)
            return user_data

    def get_ami_id(self, accelerator):
        """
        Use EKS optimized AMI since it has everything we need pre-installed
        """

        eks_owner_id = "602401143452"
        eks_amis = {
            Accelerator.NONE: "amazon-eks-node-1.14-v20190927",
            Accelerator.GPU: "amazon-eks-gpu-node-1.14-v20190927",
        }

        res = self._ec2.describe_images(
            Filters=[
                {"Name": "owner-id", "Values": [eks_owner_id]},
                {
                    "Name": "name",
                    "Values": [eks_amis.get(accelerator, Accelerator.NONE)],
                },
            ]
        )
        images = res["Images"]
        assert (
            len(images) == 1
        ), f"Multiple EKS AMIs found for {self._session.aws_region()}"
        return images[0]["ImageId"]

    def create_launch_config(
        self,
        name,
        instance_type,
        instance_role,
        user_data_template,
        security_groups=None,
        accelerator="gpu",
        max_spot_price=None,
        ebs_volume_gb=128,
        **user_data_kwargs,
    ):

        req = {
            "LaunchConfigurationName": name,
            "InstanceType": instance_type,
            "IamInstanceProfile": instance_role,
            "ImageId": self.get_ami_id(Accelerator.from_str(accelerator)),
            "SecurityGroups": security_groups,
            "AssociatePublicIpAddress": True,
            "UserData": self.get_user_data(user_data_template, **user_data_kwargs),
            "BlockDeviceMappings": [
                {
                    "DeviceName": "/dev/xvda",
                    "Ebs": {
                        "VolumeSize": ebs_volume_gb,
                        "VolumeType": "gp2",
                        "DeleteOnTermination": True,
                    },
                }
            ],
        }

        if max_spot_price:
            req["SpotMaxPrice"] = str(max_spot_price)

        log.info(f"Creating launch config: {name}")
        self._asg.create_launch_configuration(**req)

    def describe_launch_config(self, name):
        res = self._asg.describe_launch_configurations(LaunchConfigurationNames=[name])
        lcs = res["LaunchConfigurations"]
        return lcs[0] if len(lcs) == 1 else None

    def delete_launch_config(self, name):
        if self.describe_launch_config(name):
            log.info(f"Deleting asg launch config: {name}")
            self._asg.delete_launch_configuration(LaunchConfigurationName=name)

    def create_asg(self, name, size, min_size=None, max_size=None, **kwargs):
        """
        Creates an asg. For specifications on kwargs see config/sample_specs.json
        """

        if not min_size:
            min_size = size

        if not max_size:
            max_size = size

        assert min_size <= size <= max_size

        kwargs["size"] = size
        kwargs["min_size"] = min_size
        kwargs["max_size"] = max_size
        self.create_launch_config(name, **kwargs)

        log.info(f"Creating autoscaling group: {name}")
        self._asg.create_auto_scaling_group(
            AutoScalingGroupName=name,
            LaunchConfigurationName=name,
            VPCZoneIdentifier=",".join(kwargs["subnets"]),
            MinSize=min_size,
            MaxSize=max_size,
            DesiredCapacity=size,
        )

    def create_asg_sync(self, name, size, min_size=None, max_size=None, **kwargs):
        self.create_asg(name, size, min_size, max_size, **kwargs)
        _, hostnames = self.get_hostnames(name, size)
        return hostnames

    def describe_asg(self, name):
        res = self._asg.describe_auto_scaling_groups(AutoScalingGroupNames=[name])
        asgs = res["AutoScalingGroups"]
        num_asgs = len(asgs)

        return asgs[0] if num_asgs == 1 else None

    def delete_asg(self, name):
        if self.describe_asg(name):
            log.info(f"Deleting autoscaling group: {name}")
            self._asg.delete_auto_scaling_group(
                AutoScalingGroupName=name, ForceDelete=True
            )

            for _ in wait_for(f"instances in {name} to terminate"):
                if not self.describe_asg(name):
                    log.info(f"Deleted autoscaling group: {name}")
                    break

        # launch config needs to be deleted after asg
        self.delete_launch_config(name)

    def list_hostnames(self, name):
        return self.get_hostnames(name, 1)

    def get_hostnames(self, name, size):
        """
        Waits until the asg has at least <size> instances in "InService"
        state and returns their public dns names.
        """
        for _ in wait_for(f"autoscaling group: {name} to reach size >= {size}"):
            asg_desc = self.describe_asg(name)
            if not asg_desc:
                return []
            else:
                instances = asg_desc["Instances"]
                ready_instance_ids = [
                    e["InstanceId"]
                    for e in instances
                    if e["LifecycleState"] == "InService"
                ]
                if len(ready_instance_ids) >= size:
                    paginator = self._ec2.get_paginator("describe_instances")

                    hostnames = []
                    instance_ids = []
                    for e in paginator.paginate(InstanceIds=ready_instance_ids):
                        for r in e["Reservations"]:
                            for i in r["Instances"]:
                                hostnames.append(i["PublicDnsName"])
                                instance_ids.append(i["InstanceId"])
                    return instance_ids, hostnames



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import sys
import time


def wait_for(msg, timeout=300, interval=1, print_spinner=True):
    """
    for _ in wait_for("asg to provision", timeout_sec, interval_sec):
        if check_condition():.
            break
    """
    spin = ["-", "/", "|", "\\", "-", "/", "|", "\\"]
    idx = 0
    start = time.time()
    max_time = start + timeout
    while True:
        if print_spinner:
            elapsed = time.time() - start
            print(
                f"Waiting for {msg}"
                f" ({elapsed:03.0f}/{timeout:3.0f}s elapsed) {spin[idx]}\r",
                end="",
            )
            sys.stdout.flush()
            idx = (idx + 1) % len(spin)

        if time.time() >= max_time:
            raise RuntimeError(f"Timed out while waiting for: {msg}")
        else:
            time.sleep(interval)
            yield



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
from typing import Any, Dict

import torch.multiprocessing as mp
from torchelastic.agent.server.api import (
    MonitorResult,
    SimpleElasticAgent,
    Worker,
    WorkerGroup,
    WorkerSpec,
    WorkerState,
)
from torchelastic.metrics.api import prof


log = logging.getLogger(__name__)


class _DistInfo:
    """
    Container for information required to create a torch process group.
    To be created on the agent's process and passed to the worker sub-process.
    Hence this object needs to be a pure data object with no state and
    preferably only primitive member variables
    """

    __slots__ = [
        "rank",
        "group_rank",
        "local_world_size",
        "world_size",
        "master_addr",
        "master_port",
        "restart_count",
        "max_restarts",
    ]

    def __init__(
        self,
        rank: int,
        group_rank: int,
        local_world_size: int,
        world_size: int,
        master_addr: str,
        master_port: int,
        restart_count: int,
        max_restarts: int,
    ):
        self.rank = rank
        self.group_rank = group_rank
        self.local_world_size = local_world_size
        self.world_size = world_size
        self.master_addr = master_addr
        self.master_port = master_port
        self.restart_count = restart_count
        self.max_restarts = max_restarts


def _wrap(local_rank, ret_vals, dist_infos, fn, args):
    info = dist_infos[local_rank]
    os.environ["LOCAL_RANK"] = str(local_rank)
    os.environ["RANK"] = str(info.rank)
    os.environ["GROUP_RANK"] = str(info.group_rank)
    os.environ["LOCAL_WORLD_SIZE"] = str(info.local_world_size)
    os.environ["WORLD_SIZE"] = str(info.world_size)
    os.environ["MASTER_ADDR"] = info.master_addr
    os.environ["MASTER_PORT"] = str(info.master_port)
    os.environ["TORCHELASTIC_RESTART_COUNT"] = str(info.restart_count)
    os.environ["TORCHELASTIC_MAX_RESTARTS"] = str(info.max_restarts)
    ret = fn(*args)
    ret_vals[info.rank] = ret


class LocalElasticAgent(SimpleElasticAgent):
    """
    An implementation of :py:class:`torchelastic.agent.server.ElasticAgent`
    that handles host-local workers.
    This agent is deployed per host and is configured to spawn ``n`` workers.
    When using GPUs, ``n`` maps to the number of GPUs available on the host.

    The local agent does not communicate to other local agents deployed on
    other hosts, even if the workers may communicate inter-host. The worker id
    is interpreted to be a local process. The agent starts and stops all worker
    processes as a single unit.

    The worker function and argument passed to the worker function must be
    python multiprocessing compatible. To pass multiprocessing data structures
    to the workers you may create the data structure in the same multiprocessing
    context as the specified ``start_method`` and pass it as a function argument.

    Example

    ::

        def trainer(shared_queue):
            pass

        def main():
            start_method="spawn"
            shared_queue= multiprocessing.get_context(start_method).Queue()
            spec = WorkerSpec(
                        role="trainer",
                        local_world_size=nproc_per_process,
                        fn=trainer,
                        args=(shared_queue,),
                        ...<OTHER_PARAMS...>)
            agent = LocalElasticAgent(spec, start_method)
            agent.run()
    """

    def __init__(self, spec: WorkerSpec, start_method="spawn"):
        super().__init__(spec)
        self._start_method = start_method
        # pyre-fixme[8]: Attribute has type `ProcessContext`; used as `None`.
        self._process_context: mp.ProcessContext = None
        # a map that holds return values for each worker fn
        # ret_val[0] holds the return value for worker_0 (global rank 0)
        self._manager = mp.get_context(start_method).Manager()
        self._ret_vals = self._manager.dict()

    @prof
    def _stop_workers(self, worker_group: WorkerGroup) -> None:
        for proc in self._process_context.processes:
            if proc.is_alive():
                proc.terminate()

    @prof
    def _start_workers(self, worker_group: WorkerGroup) -> Dict[int, Any]:
        spec = worker_group.spec
        store = worker_group.store
        master_addr, master_port = super()._get_master_addr_port(store)
        restart_count = spec.max_restarts - self._remaining_restarts

        dist_infos: Dict[int, _DistInfo] = {}
        for worker in worker_group.workers:
            local_rank = worker.local_rank
            dist_infos[local_rank] = _DistInfo(
                worker.global_rank,
                worker_group.group_rank,
                worker_group.spec.local_world_size,
                worker.world_size,
                master_addr,
                master_port,
                restart_count,
                spec.max_restarts,
            )

        self._ret_vals.clear()
        self._process_context = mp.start_processes(
            fn=_wrap,
            args=(self._ret_vals, dist_infos, spec.fn, spec.args),
            nprocs=spec.local_world_size,
            join=False,
            daemon=False,
            start_method=self._start_method,
        )

        return {
            local_rank: pid
            for local_rank, pid in enumerate(self._process_context.pids())
        }

    @prof
    def _monitor_workers(self, worker_group: WorkerGroup) -> MonitorResult:
        role = worker_group.spec.role

        # torch process context join() isn't really a join in the
        # traditional sense, it returns True if all the workers have
        # successfully finished, False if some/all are still running
        # and throws an Exception if some/all of them failed
        # passing timeout < 0 means check worker status and return immediately

        worker_pids = {w.id for w in worker_group.workers}
        pc_pids = set(self._process_context.pids())
        if worker_pids != pc_pids:
            log.error(f"[{role}] worker pids do not match process_context pids")
            return MonitorResult(WorkerState.UNKNOWN)

        try:
            if self._process_context.join(timeout=-1):
                # copy ret_vals since we do not want to return an mp map
                return MonitorResult(WorkerState.SUCCEEDED, dict(self._ret_vals))
            else:
                return MonitorResult(WorkerState.HEALTHY)
        except Exception as e:
            log.exception(f"[{role}] Worker group failed")
            return MonitorResult(
                WorkerState.FAILED,
                exceptions={w.global_rank: e for w in worker_group.workers},
            )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import datetime
import json
import logging
import random
import sys
import threading
import time
from base64 import b64decode, b64encode
from typing import Optional

import etcd
from torch.distributed import Store, TCPStore, register_rendezvous_handler
from torchelastic.rendezvous import (
    RendezvousClosedException,
    RendezvousHandler,
    RendezvousNonRetryableError,
    RendezvousTimeoutException,
)


_log_fmt = logging.Formatter("%(levelname)s %(asctime)s %(message)s")
_log_handler = logging.StreamHandler(sys.stderr)
_log_handler.setFormatter(_log_fmt)

log = logging.getLogger(__name__)
log.propagate = False
log.setLevel(logging.INFO)
log.addHandler(_log_handler)


# Retryable failure exception means the we were too late to make
# a desired state transition (e.g. because of a race condition),
# and should now restart from the beginning.
# A small delay is recommended to avoid spamming Etcd.
class EtcdRendezvousRetryableFailure(Exception):
    pass


# Similar to retryable failure, but the new state we observed suggests we
# can re-try immediately, i.e. without a need for "safety delay".
class EtcdRendezvousRetryImmediately(Exception):
    pass


# Default overall timeout for rendezvous barrier.
CONST_DEFAULT_OVERALL_TIMEOUT = 600

# Additional waiting amount after reaching num_min_workers,
# for the case rendezvous is elastic (min != max):
CONST_DEFAULT_LAST_CALL_TIMEOUT = 30

# Various constants used internally in EtcdRendezvous
CONST_ETCD_SETUP_TTL = 5
CONST_ETCD_FROZEN_TTL = 10
CONST_ETCD_JOINABLE_EPHEMERAL_TTL = 10

# Ephemeral node TTL for worker's keep-alive key:
CONST_WORKER_KEEPALIVE_TTL = 10

# TTL for the ephemeral run_id-specific directory. All rendezvous state data
# for a specific run_id (job instance) is contained within directory.
# Its only role is to clean-up rendezvous data from old runs (for the case when
# etcd server is persistent), and has no affect on correctnes, but should be
# larger than any timeouts that a worker process is expected to survive:
CONST_RUNID_SUBROOT_TTL = 7200  # 2 hours


# Delay (sleep) for a small random amount to reduce CAS failures.
# This does not affect correctness, but will reduce requests to etcd server.
def cas_delay():
    time.sleep(random.uniform(0, 0.1))


class EtcdRendezvousHandler(RendezvousHandler):
    """
    Implements a :py:class:`torchelastic.rendezvous.RendezvousHandler`
    interface backed by
    :py:class:`torchelastic.rendezvous.etcd_rendezvous.EtcdRendezvous`.

    Torchelastic uses a URL to configure the type of rendezvous to use and
    to pass implementation specific configurations to the rendezvous module.
    The basic etcd rendezvous configuration URL looks like the following
    ::

     etcd://<etcd_address>:<port>/<job_id>?min_workers=<min_workers>&max_workers=<max_workers> # noqa W605

     -- example --

     etcd://localhost:2379/1234?min_workers=1&max_workers=3

    The URL above is interpreted as follows:

    1. Use the rendezvous handler that is registered with the ``etcd``
       scheme
    2. The ``etcd`` endpoint to use is ``localhost:2379``
    3. ``job_id == 1234`` is used as the prefix in etcd (this allows one to
       share a common etcd server for multiple jobs so long as the
       ``job_ids`` are guaranteed to be unique). Note that the job id can be
       any string (e.g. does not need to be a number) as long as it is
       unique.
    4. ``min_workers=1`` and ``max_workers=3`` specifies a range for
       membership size - torchelastic starts running the job as long as the
       cluster size is greater than or equal to ``min_workers`` and admits
       up to ``max_workers`` into the cluster.

    Below are a full list of the parameters that can be passed to etcd
    rendezvous:

    +--------------------------------------------+--------------------------+
    | Parameter                                  | Description              |
    +============================================+==========================+
    | min_workers                                | minimum number of        |
    |                                            | workers for the          |
    |                                            | rendezvous to be valid   |
    +--------------------------------------------+--------------------------+
    | max_workers                                | maximum number of        |
    |                                            | workers to admit         |
    +--------------------------------------------+--------------------------+
    | timeout                                    | total timeout within     |
    |                                            | which next_rendezvous is |
    |                                            | expected to succeed      |
    |                                            | (default 600s)           |
    +--------------------------------------------+--------------------------+
    | last_call_timeout                          | additional wait amount   |
    |                                            | (“last call”) after min  |
    |                                            | number of workers has    |
    |                                            | been reached (defaults   |
    |                                            | to 30s)                  |
    +--------------------------------------------+--------------------------+
    | etcd_prefix                                | path prefix (from etcd   |
    |                                            | root), inside which all  |
    |                                            | etcd nodes will be       |
    |                                            | created (defaults to     |
    |                                            | ``/torchelastic/p2p``)   |
    +--------------------------------------------+--------------------------+
    """

    def __init__(self, rdzv_impl):
        self._rdzv_impl = rdzv_impl

    def __del__(self):
        # TODO: look into using weakref here instead.
        del self._rdzv_impl

    def next_rendezvous(self):
        rdzv_version, rank, world_size = self._rdzv_impl.rendezvous_barrier()

        log.info("Creating EtcdStore as the c10d::Store implementation")
        store = self._rdzv_impl.setup_kv_store(rdzv_version)

        return store, rank, world_size

    def is_closed(self):
        try:
            _, state = self._rdzv_impl.get_rdzv_state()
            return state["status"] == "closed"
        except etcd.EtcdKeyNotFound:
            # No rendezvous state, so it cannot be closed.
            return False

    def set_closed(self):
        self._rdzv_impl.set_closed()

    def num_nodes_waiting(self):
        try:
            _, state = self._rdzv_impl.get_rdzv_state()
            if state["status"] == "final":
                return state["num_workers_waiting"]
        except etcd.EtcdKeyNotFound:
            pass
        return 0


# TODO: we should probably handle a few additional errors,
# like EtcdLeaderElectionInProgress and EtcdWatcherCleared. These are
# only relevant for multi-node Etcd ensemble. A simple retry would work,
# but is verbose to add everywhere. Consider wrapping the client calls
# into auto-retry for these errors?
#
class EtcdRendezvous(object):
    """
    A rendezvous implementation that uses `etcd <https://etcd.io/>`__ as
    the backend store.
    """

    def __init__(
        self,
        endpoints,
        prefix,
        run_id,
        num_min_workers,
        num_max_workers,
        timeout,
        last_call_timeout,
        **kwargs,
    ):
        self._prefix = prefix
        self._run_id = run_id
        self._num_min_workers = num_min_workers
        self._num_max_workers = num_max_workers
        self._timeout = timeout
        self._last_call_timeout = last_call_timeout

        # For cleaning up TTL refresher threads (for ephemeral keys)
        self._lease_run_id_stop = None
        self._lease_this_rank_stop = None

        if not self._prefix.endswith("/"):
            self._prefix += "/"

        self.client = etcd.Client(host=endpoints, allow_reconnect=True, **kwargs)
        log.info("Etcd machines: " + str(self.client.machines))

        # Setup a permanent prefix dir, if didn't exist
        if self._prefix != "/":
            self.create_path_if_not_exists(self._prefix)

        # Lease a "sub-root" node specific to this job instance (run_id)
        self.create_path_if_not_exists(self.get_path(""), ttl=CONST_RUNID_SUBROOT_TTL)
        self._lease_run_id_stop = self.setup_lease_renewal(
            self.get_path(""), ttl=CONST_RUNID_SUBROOT_TTL
        )

        # Subdir for all rendezvous work
        self.create_path_if_not_exists(self.get_path("/rdzv"))

        # Create a rendezvous version counter, if doesn't exist
        try:
            self.client.write(
                key=self.get_path("/rdzv/version_counter"), value="0", prevExist=False
            )
        except etcd.EtcdAlreadyExist:
            pass

    def __del__(self):
        # TODO: look into using weakref here instead.
        if self._lease_run_id_stop is not None:
            self._lease_run_id_stop.set()

        if self._lease_this_rank_stop is not None:
            self._lease_this_rank_stop.set()

    #
    # Main entry point for next rendezvous.
    # This method is blocking until rendezvous succeeds or a timeout occurs.
    #
    def rendezvous_barrier(self):
        self._rendezvous_deadline = time.time() + self._timeout
        while True:
            if time.time() > self._rendezvous_deadline:
                raise RendezvousTimeoutException()

            log.info("Attempting to join next rendezvous")
            try:
                # Dis-own our lease in the previous rendezvous, if exists
                if self._lease_this_rank_stop is not None:
                    self._lease_this_rank_stop.set()

                return self.init_phase()

            except EtcdRendezvousRetryImmediately:
                # The type of failure suggests we can retry without delay
                pass

            except EtcdRendezvousRetryableFailure:
                # In case of retryable failure, wait a small delay
                # to avoid spamming etcd
                time.sleep(1)

            except RendezvousTimeoutException:
                log.info("Rendezvous timeout occured in EtcdRendezvousHandler")
                raise

            except RendezvousClosedException:
                log.info(
                    f"Rendezvous for run_id={self._run_id} was observed to be closed"
                )
                raise

            except RendezvousNonRetryableError:
                raise

            except Exception as e:
                # In case of a general exception, wait a small delay
                # to avoid spamming etcd
                # FIXME: there are a few things that fall under this like
                # etcd.EtcdKeyNotFound, etc, which could be handled more explicitly.
                log.info("Rendezvous attempt failed, will retry. Reason: " + str(e))
                time.sleep(1)

    #
    # Initially, the rendezvous state is expected to be one of:
    #  a) empty (non-existent) - in this case we try to create a new one.
    #  b) joinable - we try to join it.
    #  c) final - we announce ourselves as waiting, and go into monitoring mode
    #             for this rendezvous.
    #
    # Any other state is considered transitional, and will be retried after
    # a short delay.
    #
    def init_phase(self):
        try:
            active_version = self.try_create_rendezvous()
            state = json.loads(active_version.value)
            log.info("New rendezvous state created: " + str(state))
        except etcd.EtcdAlreadyExist:
            active_version, state = self.get_rdzv_state()
            # Note: it is possible for above query to fail (etcd.EtcdKeyNotFound),
            # but this is ok for us - just means we'll restart from beginning.
            log.info("Observed existing rendezvous state: " + str(state))

        if state["status"] == "closed":
            raise RendezvousClosedException()

        if state["status"] == "joinable":
            return self.join_phase(state["version"])

        if state["status"] == "final":
            self.handle_existing_rendezvous(state["version"])
            raise EtcdRendezvousRetryImmediately()

        # We observed some intermediate state, which is best handled by retrying later
        self.try_wait_for_state_change(etcd_index=active_version.etcd_index + 1)
        raise EtcdRendezvousRetryableFailure()

    #
    # We observed a rendezvous state in 'joinable' state, and attempt to join this
    # particular version, and then wait for all other peers to join.
    #
    def join_phase(self, expected_version):
        # Failure to join will propagate an exception, causing a re-entry.
        active_version, this_rank = self.join_rendezvous(expected_version)
        state = json.loads(active_version.value)
        log.info(
            "Joined rendezvous version {} as rank {}. Full state: {}".format(
                state["version"], this_rank, state
            )
        )

        # If this worker was first to reach num_min_workers requirement,
        # and rendezvous is still joinable (therefore it is elastic),
        # then this worker will be repsonsible for waiting out the "last call"
        # timeout and closing (i.e. transitioning to 'frozen') the rendezvous
        # afterwards.
        # As a safety against a potential failure of this worker (during the
        # last call timeout), the rendezvous state is made ephemeral
        # when min_num_workers is reached.

        if this_rank == self._num_min_workers - 1 and state["status"] == "joinable":
            log.info("Rank {} is responsible for join last call.".format(this_rank))
            last_call_deadline = time.time() + self._last_call_timeout
            self.handle_join_last_call(expected_version, last_call_deadline)
            log.info("Rank {} finished join last call.".format(this_rank))

        # Wait for rendezvous state to be frozen, which means a fixed set of peers
        log.info("Waiting for remaining peers.")
        active_version = self.wait_for_peers(expected_version)
        state = json.loads(active_version.value)

        assert (
            state["version"] == expected_version
        ), "Logic error: failed to observe version mismatch"

        return self.confirm_phase(expected_version, this_rank)

    #
    # Once the rendezvous state transitiosn from 'joinable' to 'frozen',
    # we have every participant confirm their membership and setup per-member
    # keep-alive TTL keys, and then wait for all other participants to confirm,
    # which would then successfully conclude this rendezvous.
    #
    def confirm_phase(self, expected_version, this_rank):
        log.info("All peers arrived. Confirming membership.")
        self.confirm_membership(expected_version, this_rank)

        log.info("Waiting for confirmations from all peers.")
        active_version = self.wait_for_final(expected_version)
        state = json.loads(active_version.value)

        log.info(
            "Rendezvous version {} is complete. Final state: {}".format(
                state["version"], state
            )
        )

        # Rendezvous version number; our rank in it; world size
        return state["version"], this_rank, len(state["participants"])

    #
    # Handle the case when there's an existing (state 'final) rendezvous already
    # in place, and we have to announce ourselves waiting, and wait until
    # the next rendezvous opportunity.
    #
    def handle_existing_rendezvous(self, expected_version):
        # If statie is 'final' -> increment num_workers_waiting
        # Then, observe state changes:
        #   1. if it's no longer final -> bail out and re-try
        #   2. if keep alives are missing, destroy it and bail out.
        active_state = self.announce_self_waiting(expected_version)
        log.info(
            "Added self to waiting list. Rendezvous full state: {}".format(
                active_state.value
            )
        )

        self.wait_for_rendezvous_to_free(expected_version)
        log.info("Previously existing rendezvous state changed. Will re-try joining.")

    # Create new rendezvous state, or raise etcd.EtcdAlreadyExist exception.
    def try_create_rendezvous(self):
        # Initially active_version is ephemeral - this is to handle the
        # possibility that might fail to complete the setup transaction,
        # i.e. the transition "setup" -> "joinable".
        active_version = self.client.write(
            key=self.get_path("/rdzv/active_version"),
            value=json.dumps({"status": "setup"}),
            prevExist=False,
            ttl=CONST_ETCD_SETUP_TTL,
        )

        try:
            version_counter = self.client.get(self.get_path("/rdzv/version_counter"))
            version_counter.value = str(int(version_counter.value) + 1)
            self.client.update(version_counter)
        except (etcd.EtcdKeyNotFound, etcd.EtcdCompareFailed):
            raise RendezvousNonRetryableError(
                "Unexpected state of EtcdRendezvousHandler, worker needs to die."
            )

        # Any failure below results in declaring a retryable rendezvous failure.
        # The ephemeral /rdzv/active_version will expire and someone can then
        # re-try the setup process.

        # Create directory node for participant data
        self.client.write(
            key=self.get_path("/rdzv/v_{}".format(version_counter.value)),
            value=None,
            dir=True,
            prevExist=False,
        )

        # Publish rendezvous version and signal it is ready-to-be-joined.
        # If rendezvous was set closed just before this, a retry will happen,
        # where the closed condition will be handled.
        return self.client.test_and_set(
            key=self.get_path("/rdzv/active_version"),
            value=json.dumps(
                {
                    "status": "joinable",
                    "version": version_counter.value,
                    "participants": [],
                }
            ),
            prev_value=active_version.value,
        )

    # Helper method for the join phase.
    def join_rendezvous(self, expected_version):
        # Use compare-and-swap to add self to rendezvous state:
        while True:
            cas_delay()
            active_version, state = self.get_rdzv_state()

            if state["status"] != "joinable":
                raise EtcdRendezvousRetryableFailure(
                    "Rendezvous state became non-joinable before we could join. "
                    "Must join next one."
                )

            if state["version"] != expected_version:
                raise EtcdRendezvousRetryImmediately(
                    "Rendezvous version changed. Must try join the new one."
                )

            assert (
                len(state["participants"]) < self._num_max_workers
            ), "Logic error: joinable rendezvous should always have space left"

            this_rank = len(state["participants"])
            state["participants"].append(this_rank)

            # When reaching min workers, or changing state to frozen, we'll set
            # the active_version node to be ephemeral.
            if len(state["participants"]) == self._num_max_workers:
                state["status"] = "frozen"
                state["keep_alives"] = []
                set_ttl = CONST_ETCD_FROZEN_TTL
            elif len(state["participants"]) >= self._num_min_workers:
                set_ttl = CONST_ETCD_JOINABLE_EPHEMERAL_TTL
            else:
                set_ttl = None

            try:
                # Compare-and-swap.
                active_version = self.client.test_and_set(
                    key=self.get_path("/rdzv/active_version"),
                    value=json.dumps(state),
                    prev_value=active_version.value,
                    ttl=set_ttl,
                )
                # We succeeded joining.
                return active_version, this_rank

            except etcd.EtcdCompareFailed:
                log.info("Join rendezvous CAS unsuccessful, retrying")

    # Helper method for the join phase.
    def wait_for_peers(self, expected_version):
        active_version, state = self.get_rdzv_state()
        while True:
            if state["status"] == "frozen" and state["version"] == expected_version:
                # Success, all peers arrived.
                return active_version

            elif state["status"] == "joinable" and state["version"] == expected_version:
                # Continue waiting for any interesting events.
                active_version, state = self.try_wait_for_state_change(
                    etcd_index=active_version.etcd_index + 1
                )

            else:
                # No valid transition possible at this point
                raise EtcdRendezvousRetryableFailure(
                    "Rendezvous state transition no longer possible. Must re-enter."
                )

    # Helper method for the confirm phase
    def confirm_membership(self, expected_version, this_rank):
        # Compare-and-swap loop
        while True:
            cas_delay()
            active_version, state = self.get_rdzv_state()

            if state["status"] != "frozen":
                raise EtcdRendezvousRetryImmediately(
                    "Rendezvous no longer frozen, before we confirmed. "
                    "Must join next one"
                )
            if state["version"] != expected_version:
                raise EtcdRendezvousRetryImmediately(
                    "Rendezvous version changed. Must try join the new one."
                )

            this_lease_key = self.get_path(
                "/rdzv/v_{}/rank_{}".format(expected_version, this_rank)
            )
            self.client.set(this_lease_key, value=None, ttl=CONST_WORKER_KEEPALIVE_TTL)

            state["keep_alives"].append(this_lease_key)
            if len(state["keep_alives"]) == len(state["participants"]):
                # Everyone confirmed (this rank is last to do so)
                state["status"] = "final"
                state["num_workers_waiting"] = 0
                finalize = True
            else:
                finalize = False

            try:
                # Compare-and-swap. If new state is still frozen, keep it ephemeral.
                active_version = self.client.test_and_set(
                    key=self.get_path("/rdzv/active_version"),
                    value=json.dumps(state),
                    prev_value=active_version.value,
                    ttl=None if finalize else CONST_ETCD_FROZEN_TTL,
                )

                self._lease_this_rank_stop = self.setup_lease_renewal(
                    this_lease_key, ttl=CONST_WORKER_KEEPALIVE_TTL
                )
                return active_version

            except etcd.EtcdCompareFailed:
                log.info("Confirm membership CAS unsuccessful, retrying")

    # Helper method for the confirm phase
    def wait_for_final(self, expected_version):
        active_version, state = self.get_rdzv_state()
        while True:
            if state["status"] == "final" and state["version"] == expected_version:
                # Succcess. This rendezvous is final, and we accept it.
                return active_version

            elif state["status"] == "frozen" and state["version"] == expected_version:
                # Continue waiting for any interesting events.
                active_version, state = self.try_wait_for_state_change(
                    etcd_index=active_version.etcd_index + 1
                )

            else:
                # No valid transition possible at this point
                raise EtcdRendezvousRetryableFailure(
                    "Rendezvous state transition no longer possible. Must re-enter."
                )

    # Announce this worker is waiting (via num_workers_waiting counter) to join next
    # rendezvous, but only if state and version match.
    def announce_self_waiting(self, expected_version):
        while True:
            cas_delay()
            active_version, state = self.get_rdzv_state()

            if state["status"] != "final" or state["version"] != expected_version:
                raise EtcdRendezvousRetryImmediately()

            # Increment counter to signal an additional waiting worker.
            state["num_workers_waiting"] += 1

            try:
                active_version = self.client.test_and_set(
                    key=self.get_path("/rdzv/active_version"),
                    value=json.dumps(state),
                    prev_value=active_version.value,
                )
                return active_version

            except etcd.EtcdCompareFailed:
                log.info("Announce self as waiting CAS unsuccessful, retrying")

    #
    # When there's an existing valid rendezvous in state 'final', we have to wait until
    # the next opportunity to join.
    #
    # Such opportunity may come from:
    #  a) rendezvous state changed by someone else, in which case we unblock and retry.
    #  b) rendezvous becomes invalid because at least one member failed to renew their
    #     leased keep_alive node. We detect this, and destroy the rendezvous.
    #
    def wait_for_rendezvous_to_free(self, expected_version):
        active_version, state = self.get_rdzv_state()
        while True:
            if state["status"] != "final" or state["version"] != expected_version:
                return

            # Check if current rendezvous state is valid, in the sense that all
            # its members are alive (renewing their lease).
            # If not, try destroy this rendezvous, so a new one can be created.
            alive_members = self.client.get(
                self.get_path("/rdzv/v_{version}".format(version=expected_version))
            )
            keep_alive_keys = [ch.key for ch in alive_members.children]

            for key in state["keep_alives"]:
                if key not in keep_alive_keys:
                    # This participant didn't renew their lease. We'll declare this
                    # rendezvous version as dead (but only if it hadn't changed)
                    log.info("Keep-alive key {} is not renewed.".format(key))
                    log.info(
                        "Rendevous version {} is incomplete. ".format(expected_version)
                    )
                    log.info("Attempting to destroy it.")

                    # Compare-and-delete operation. Throws if compare failed,
                    # which means rendezvous was already destroyed/re-created/closed,
                    # and we can try to re-enter the barrier.
                    self.client.delete(
                        key=self.get_path("/rdzv/active_version"),
                        prevValue=active_version.value,
                    )

                    log.info(
                        "Destroyed rendezvous version {} successfully.".format(
                            expected_version
                        )
                    )

                    # We can return (and retry) immediately
                    return

            # Existing rendezvous seems valid, no reason to destroy it.
            # We just have to wait until something changes and re-check.
            try:
                overall_timeout = (
                    max(self._rendezvous_deadline - time.time(), 0.0) + 1.0
                )
                self.client.watch(
                    key=self.get_path("/rdzv"),
                    index=active_version.etcd_index + 1,
                    recursive=True,
                    timeout=overall_timeout,
                )
            except (etcd.EtcdEventIndexCleared, etcd.EtcdWatchTimedOut):
                pass

            if time.time() > self._rendezvous_deadline:
                raise RendezvousTimeoutException()
            active_version, state = self.get_rdzv_state()

    #
    # After we reach min number of workers, one particular worker takes on the
    # responsibility of waiting an additional timeout before closing the join window.
    # If the worker responsible for this fails, the rendezvous will be destroyed due
    # to expiring TTL, and the other participants will re-rendezvous.
    #
    def handle_join_last_call(self, expected_version, deadline):
        # Here we expect to see state <joinable, expected_version>
        # Exit gracefully if either:
        #  1. state becomes <frozen, expected_version>
        #  2. timeout happens (reaching deadline), in which case
        #     we try the tranisiton to <frozen, expected_version>
        # Exit with exception otherwise.

        active_version, state = self.get_rdzv_state()
        while True:
            if state["status"] == "frozen" and state["version"] == expected_version:
                # Worker set became frozen before last-call timeout. This is possible
                # when num_max_workers is reached before the tiemout.
                return

            if state["status"] != "joinable" or state["version"] != expected_version:
                raise EtcdRendezvousRetryableFailure(
                    "Rendezvous state transition no longer possible. Must re-enter."
                )

            # If timeout occurred, attempt a state transition (joinable -> frozen)
            if time.time() >= deadline:
                state["status"] = "frozen"
                state["keep_alives"] = []
                try:
                    active_version = self.client.test_and_set(
                        key=self.get_path("/rdzv/active_version"),
                        value=json.dumps(state),
                        prev_value=active_version.value,
                        ttl=CONST_ETCD_FROZEN_TTL,
                    )
                    # We successfully made this rendezvous frozen.
                    return
                except etcd.EtcdCompareFailed:
                    log.info("Join last-call transition CAS unsuccessful. Will retry")
                    cas_delay()
                    active_version, state = self.get_rdzv_state()
                    continue

            # Timeout did not occur, so we must refresh TTL, and wait for
            # further changes. Note: we only want TTL to be refreshed if
            # state is still joinable, hence we use CAS for that here,
            # even though we don't change any of the data.
            try:
                active_version = self.client.test_and_set(
                    key=self.get_path("/rdzv/active_version"),
                    value=active_version.value,
                    prev_value=active_version.value,
                    ttl=CONST_ETCD_JOINABLE_EPHEMERAL_TTL,
                )

                # Minimize "oversleeping":
                timeout = min(
                    CONST_ETCD_JOINABLE_EPHEMERAL_TTL / 2,
                    deadline - time.time() + 1.0,  # Oversleeping by 1s is ok.
                )
                active_version, state = self.try_wait_for_state_change(
                    etcd_index=active_version.etcd_index + 1, timeout=timeout
                )
            except etcd.EtcdCompareFailed:
                log.info("Join last-call TTL refresh CAS unsuccessful, will retry")
                cas_delay()
                active_version, state = self.get_rdzv_state()

    # Mark rendezvous 'closed' for current run_id, which is used to signal other
    # participants to not attempt to perform (re-)rendezvous. This is useful
    # when one of the workers decides the job is complete.
    def set_closed(self):
        while True:
            active_version, state = self.get_rdzv_state()

            if state["status"] == "closed":
                # Already closed by someone else.
                return

            state["status"] = "closed"
            try:
                self.client.test_and_set(
                    key=self.get_path("/rdzv/active_version"),
                    value=json.dumps(state),
                    prev_value=active_version.value,
                )
                return

            except etcd.EtcdCompareFailed:
                log.info("Set closed CAS unsuccessful, retrying")
                cas_delay()

    def get_rdzv_state(self):
        active_version = self.client.get(key=self.get_path("/rdzv/active_version"))
        return active_version, json.loads(active_version.value)

    def try_wait_for_state_change(self, etcd_index, timeout=None):
        # Don't sleep past the overall deadline (at least more than by 1s)
        overall_timeout = max(self._rendezvous_deadline - time.time(), 0.0) + 1.0
        timeout = overall_timeout if timeout is None else min(timeout, overall_timeout)

        try:
            self.client.watch(
                self.get_path("/rdzv/active_version"), index=etcd_index, timeout=timeout
            )
        except (etcd.EtcdEventIndexCleared, etcd.EtcdWatchTimedOut):
            pass

        if time.time() > self._rendezvous_deadline:
            raise RendezvousTimeoutException()

        # Unfortunately, we have to do another fetch in order to get last etcd_index.
        return self.get_rdzv_state()

    def get_path(self, path):
        if not path.startswith("/"):
            path = "/" + path

        return "{prefix}run_{run_id}{path}".format(
            prefix=self._prefix, run_id=self._run_id, path=path
        )

    def create_path_if_not_exists(self, full_path, ttl=None):
        try:
            self.client.write(
                key=full_path, value=None, dir=True, prevExist=False, ttl=ttl
            )
        except etcd.EtcdAlreadyExist:
            pass

    def setup_lease_renewal(self, full_path, ttl):
        # NOTE: For ephemeral key TTL renewal (~lease) to work correctly,
        # make sure you don't call any long-blocking methods that do not
        # release the Python's GIL! An example of this is calling a pybind11
        # extension function that is blocking / long-running, but is not
        # doing a scoped release of the GIL.
        def lease_worker(client, path, ttl, stop_event):
            while True:
                try:
                    client.refresh(path, ttl=ttl)
                except etcd.EtcdKeyNotFound:
                    break

                if stop_event.wait(timeout=ttl / 2):
                    break

        lease_stop_event = threading.Event()
        lease_thread = threading.Thread(
            target=lease_worker, args=(self.client, full_path, ttl, lease_stop_event)
        )

        lease_thread.daemon = True
        lease_thread.start()

        return lease_stop_event

    def store_extra_data(self, rdzv_version, key, value):
        node = self.get_path("/rdzv/v_{}/extra_data".format(rdzv_version))
        try:
            # If first time we are storing anything:
            extra_data = self.client.write(
                key=node, value=json.dumps({key: value}), prevExist=False
            )
            return
        except etcd.EtcdAlreadyExist:
            pass

        # CAS loop, to make sure we don't lose concurrent stores.
        while True:
            # We never delete extra_data. Failure here should be fatal, no special handling.
            extra_data = self.client.get(node)

            new_extra_data_value = json.loads(extra_data.value)
            new_extra_data_value[key] = value

            try:
                extra_data = self.client.test_and_set(
                    key=node,
                    value=json.dumps(new_extra_data_value),
                    prev_value=extra_data.value,
                )
                return
            except etcd.EtcdCompareFailed:
                log.info("Store extra_data CAS unsuccessful, retrying")
                time.sleep(0.1)

    def load_extra_data(self, rdzv_version, key, timeout=None):
        # 'extra_data' node itself, and the directory it is located in:
        node = self.get_path("/rdzv/v_{}/extra_data".format(rdzv_version))
        node_dir = self.get_path("/rdzv/v_{}".format(rdzv_version))

        # TODO: implement timeout
        # https://github.com/pytorch/elastic/issues/12
        while True:
            # Combined wait for the node itself, and the key inside it.
            root = self.client.get(node_dir)

            # Find the extra_data node, if it exists
            extra_data = [n for n in root.children if n.key == node]
            assert len(extra_data) <= 1

            # Node for extra_data exists, check the desired key inside it.
            if len(extra_data) == 1:
                extra_data_dict = json.loads(extra_data[0].value)
                if key in extra_data_dict:
                    return extra_data_dict[key]

            # The 'extra_data' node doesn't exist, or they key isn't published yet.
            # Wait for interesting events on the extra_data node and retry.
            try:
                self.client.watch(node, index=root.etcd_index + 1)
            except (etcd.EtcdEventIndexCleared, etcd.EtcdWatchTimedOut):
                pass

    def setup_kv_store(self, rdzv_version):
        store_path = self.get_path(f"/rdzv/v_{rdzv_version}/kv")
        self.create_path_if_not_exists(store_path)
        return EtcdStore(etcd_client=self.client, etcd_store_prefix=store_path)


# pyre-fixme[11]: Annotation `Store` is not defined as a type.
class EtcdStore(Store):
    """
    Implements a c10 Store interface by piggybacking on the rendezvous etcd instance.
    """

    def __init__(
        self,
        etcd_client,
        etcd_store_prefix,
        timeout: Optional[datetime.timedelta] = None,
    ):
        super().__init__()  # required for pybind trampoline.

        self.client = etcd_client
        self.prefix = etcd_store_prefix

        # Default timeout same as in c10d/Store.hpp
        self.timeout = (
            timeout if timeout is not None else datetime.timedelta(seconds=300)
        )
        if not self.prefix.endswith("/"):
            self.prefix += "/"

    def set(self, key, value):
        """
        Write a key/value pair into EtcdStore.
        Both key and value may be either Python str or bytes.
        """
        self.client.set(key=self.prefix + self._encode(key), value=self._encode(value))

    def get(self, key) -> bytes:
        """
        Get a value by key, possibly doing a blocking wait.

        If key is not immediately present, will do a blocking wait
        for at most `timeout` duration or until the key is published.
        If key still not published after timeout, a LookupError is raised.

        Return value is of type bytes.
        """
        b64_key = self.prefix + self._encode(key)
        kvs = self._try_wait_get([b64_key])

        if kvs is None:
            raise LookupError(f"Key {key} not found in EtcdStore")

        return self._decode(kvs[b64_key])

    def add(self, key, num: int) -> int:
        """
        Atomically increment a value by an integer amount, returns the new value.
        Returned value is of type `int`.

        The integer is represented as a string using base 10.
        If key is not present, a default value of "0" will be assumed.
        """
        b64_key = self._encode(key)
        # c10d Store assumes value is an integer represented as a decimal string
        try:
            # Assume default value "0", if this key didn't yet:
            node = self.client.write(
                key=self.prefix + b64_key,
                value=self._encode(str(num)),  # i.e. 0 + num
                prevExist=False,
            )
            return int(self._decode(node.value))
        except etcd.EtcdAlreadyExist:
            pass

        while True:
            # Note: c10d Store does not have a method to delete keys, so we
            # can be sure it's still there.
            node = self.client.get(key=self.prefix + b64_key)
            new_value = self._encode(str(int(self._decode(node.value)) + num))
            try:
                node = self.client.test_and_set(
                    key=node.key, value=new_value, prev_value=node.value
                )
                return int(self._decode(node.value))
            except etcd.EtcdCompareFailed:
                cas_delay()

    def wait(self, keys, override_timeout: Optional[datetime.timedelta] = None):
        """
        Waits until all of the keys are published, or until timeout.
        If timeout occurs first, a LookupError is raied.
        """
        b64_keys = [self.prefix + self._encode(key) for key in keys]
        kvs = self._try_wait_get(b64_keys, override_timeout)
        if kvs is None:
            raise LookupError("Timeout while waiting for keys in EtcdStore")
        # No return value on success

    def check(self, keys) -> bool:
        """
        Check if all of the keys are immediately present (without waiting).
        Returns bool.
        """
        b64_keys = [self.prefix + self._encode(key) for key in keys]
        kvs = self._try_wait_get(
            b64_keys,
            override_timeout=datetime.timedelta(microseconds=1),  # as if no wait
        )
        return kvs is not None

    def set_timeout(self, timeout: datetime.timedelta):
        """
        Change the timeout used for all future operations.
        """
        self.timeout = timeout

    #
    # Encode key/value data in base64, so we can store arbitrary binary data
    # in EtcdStore. Input can be `str` or `bytes`.
    # In case of `str`, utf-8 encoding is assumed.
    #
    def _encode(self, value) -> str:
        if type(value) == bytes:
            return b64encode(value).decode()
        elif type(value) == str:
            return b64encode(value.encode()).decode()
        raise ValueError("Value must be of type str or bytes")

    #
    # Decode a base64 string (of type `str` or `bytes`).
    # Return type is `bytes`, which is more convenient with the Store interface.
    #
    def _decode(self, value) -> bytes:
        if type(value) == bytes:
            return b64decode(value)
        elif type(value) == str:
            return b64decode(value.encode())
        raise ValueError("Value must be of type str or bytes")

    #
    # Get all of the (base64-encoded) etcd keys at once, or wait until all the keys
    # are published or timeout occurs.
    # This is a helper method for the public interface methods.
    #
    # On success, a dictionary of {etcd key -> etcd value} is returned.
    # On timeout, None is returned.
    #
    def _try_wait_get(self, b64_keys, override_timeout=None):
        timeout = self.timeout if override_timeout is None else override_timeout
        deadline = time.time() + timeout.total_seconds()

        while True:
            # Read whole directory (of keys), filter only the ones waited for
            all_nodes = self.client.get(key=self.prefix)
            req_nodes = {
                node.key: node.value
                for node in all_nodes.children
                if node.key in b64_keys
            }

            if len(req_nodes) == len(b64_keys):
                # All keys are available
                return req_nodes

            watch_timeout = deadline - time.time()
            if watch_timeout <= 0:
                return None

            try:
                self.client.watch(
                    key=self.prefix,
                    recursive=True,
                    timeout=watch_timeout,
                    index=all_nodes.etcd_index + 1,
                )
            except etcd.EtcdWatchTimedOut:
                if time.time() >= deadline:
                    return None
                else:
                    continue
            except etcd.EtcdEventIndexCleared:
                continue


def _get_socket_with_port():
    import socket

    addrs = socket.getaddrinfo(
        host="localhost", port=None, family=socket.AF_UNSPEC, type=socket.SOCK_STREAM
    )
    for addr in addrs:
        family, type, proto, _, _ = addr
        try:
            s = socket.socket(family, type, proto)
            s.bind(("localhost", 0))
            s.listen(0)
            return s
        except OSError as e:
            s.close()
            log.info("Socket creation attempt failed: " + e)

    raise RuntimeError("Failed to create a socket")


# Helper for _etcd_rendezvous_handler(url)
def _parse_etcd_client_params(params):
    kwargs = {}
    if "protocol" in params:
        protocol = params["protocol"]
        assert protocol in ["http", "https"], "Protocol must be http or https."
        kwargs["protocol"] = protocol
    if "cacert" in params:
        kwargs["ca_cert"] = params["cacert"]
    if "cert" in params:
        if "key" in params:
            # python-etcd client expects key as a second element of `cert` tuple
            kwargs["cert"] = (params["cert"], params["key"])
        else:
            kwargs["cert"] = params["cert"]
    return kwargs


# Handler for torch.distributed "static" registration
def _etcd_rendezvous_handler(url):
    """
    Example URLs:
        etcd://localhost:2379/123?min_workers=4&max_workers=8&timeout=300
        etcd://192.168.0.42/123?etcd_prefix=/custom_prefix/foo&min_workers=4
        etcd://localhost:2379/123?min_workers=4&protocol=https&cacert=/etc/kubernetes/certs/ca.crt&cert=/etc/kubernetes/certs/client.crt&key=/etc/kubernetes/certs/client.key

    Where:
        123 - the run_id (unique id for this training job instance),
        min_workers=4 - min number of workers expected to join the rendezvous,
        max_workers=8 - max number of workers allowed to join the rendezvous,
                        defaults to min_workers is not specified.
        timeout=300 - total timeout within which next_rendezvous is expected to
                      succeed; a RendezvousTimeoutException is raised otherwise;
                      Defaults is 600 (10 minutes).
        last_call_timeout - additional wait amount ("last call") after
                            min number of workers has been reached.
                            Defaults to 30 seconds.
        etcd_prefix - path prefix (from etcd root), inside which all
                      etcd nodes will be created.
                      Default is "/torchelastic/p2p".
        protocol=https - http (default) or https to access etcd.
        cacert=/etc/kubernetes/certs/ca.crt - CA cert to access etcd,
                    only makes sense with https.
        cert=/etc/kubernetes/certs/client.crt - client cert to access etcd,
                    only makes sense with https.
        key=/etc/kubernetes/certs/client.key - client key to access etcd,
                    only makes sense with https.

    """
    import re
    from urllib.parse import urlparse

    url = urlparse(url)
    assert url.scheme == "etcd"

    # Etcd endpoints. (Current url format only allows a single host)
    endpoint = url.netloc
    match = re.match(r"(.+):(\d+)$", endpoint)  # check if port was provided
    if match:
        etcd_endpoints = ((match.group(1), int(match.group(2))),)
    else:
        # Use default etcd port
        etcd_endpoints = ((endpoint, 2379),)

    # Run ID value -> unique identifier of this training job instance:
    # typically a job_id or name assigned by the scheduler or user
    run_id = url.path.strip("/")

    # Parse all of query parameters:
    params = dict(pair.split("=") for pair in filter(None, url.query.split("&")))

    etcd_prefix = params.get("etcd_prefix", "/torchelastic/p2p")
    num_min_workers = int(params["min_workers"])
    num_max_workers = int(params.get("max_workers", num_min_workers))
    assert num_min_workers >= 1, "Min number of workers should be at least 1"
    assert (
        num_max_workers >= num_min_workers
    ), "Max number of workers cannot be less than min number of workers"

    timeout = int(params.get("timeout", CONST_DEFAULT_OVERALL_TIMEOUT))
    last_call_timeout = int(
        params.get("last_call_timeout", CONST_DEFAULT_LAST_CALL_TIMEOUT)
    )

    kwargs = _parse_etcd_client_params(params)

    # Etcd rendezvous implementation
    etcd_rdzv = EtcdRendezvous(
        endpoints=etcd_endpoints,
        prefix=etcd_prefix,
        run_id=run_id,
        num_min_workers=num_min_workers,
        num_max_workers=num_max_workers,
        timeout=timeout,
        last_call_timeout=last_call_timeout,
        **kwargs,
    )
    return EtcdRendezvousHandler(rdzv_impl=etcd_rdzv)


# torchelastic.rendezvous.RendezvousHandler using etcd (API v2):
register_rendezvous_handler("etcd", _etcd_rendezvous_handler)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import getpass
import json
import logging
import os
import sys
from os.path import expanduser
from urllib.parse import urlparse

import auth
from autoscaling import AutoScalingGroup
from cloudformation import CloudFormation
from s3 import S3


log = logging.getLogger(__name__)
PETCTL_DIR = os.path.join(expanduser("~"), ".petctl")
PETCTL_CONFIG_FILE = os.path.join(PETCTL_DIR, "config")
SPECS_FILE = os.path.join(PETCTL_DIR, "specs.json")


def split_args(args, delimiter="--"):
    if delimiter in args:
        idx = args.index(delimiter)
        if idx == (len(args) - 1):
            return args, []
        else:
            return args[0:idx], args[idx + 1 :]
    else:
        return args, []


def parse_arguments(args, **default_args):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--specs_file",
        help="see https://github.com/pytorch/elastic/blob/master/aws/README.md#create-specs-file",  # noqa B950
    )
    parser.set_defaults(**default_args)

    subparser = parser.add_subparsers(
        title="actions", description="run_job | kill_job", dest="command"
    )

    # -----------------------------------------
    # Run Job
    # -----------------------------------------
    parser_run_job = subparser.add_parser(
        "run_job", help="runs a torchelastic job on asg"
    )
    parser_run_job.add_argument("--name", required=True, help="name of the job")
    parser_run_job.add_argument(
        "--min_size",
        type=int,
        required=False,
        help="minimum number of worker hosts to continue training",
    )
    parser_run_job.add_argument(
        "--max_size",
        type=int,
        required=False,
        help="maximum number of worker hosts to allow scaling out",
    )
    parser_run_job.add_argument(
        "--size",
        type=int,
        required=True,
        help="number  of worker hosts to start the job with",
    )
    parser_run_job.add_argument(
        "--instance_type", required=False, help="Instance type to run the job on"
    )
    parser_run_job.add_argument(
        dest="script_path",
        help="script or script dir path (e.g. ~/script.py, s3://..., docker://)",
    )
    parser_run_job.set_defaults(func=run_job)

    # -----------------------------------------
    # Kill Job
    # -----------------------------------------
    parser_kill_job = subparser.add_parser(
        "kill_job", help="kills a torchelastic job on asg"
    )

    parser_kill_job.add_argument(dest="job_name", help="name of the job to kill")

    parser_kill_job.set_defaults(func=kill_job)

    # -----------------------------------------
    # List hosts in job
    # -----------------------------------------
    parser_list_hosts = subparser.add_parser(
        "list_hosts", help="lists InService hosts in the job"
    )
    parser_list_hosts.add_argument(
        dest="job_name", help="name of the job to list the hosts for"
    )
    parser_list_hosts.set_defaults(func=list_hosts)

    # -----------------------------------------
    # Upload script
    # -----------------------------------------
    parser_upload = subparser.add_parser("upload", help="uploads the file/dir to s3")
    parser_upload.add_argument(
        dest="script_path",
        help="script or script dir path (e.g. ~/script.py, s3://..., docker://)",
    )
    parser_upload.add_argument(
        dest="s3_dest",
        help="s3 destination (default: s3://{s3_bucket}/{s3_prefix}/{USER}/scripts)",
    )
    parser_upload.set_defaults(func=upload_script)

    # -----------------------------------------
    # Configure
    # -----------------------------------------
    subparser.add_parser("configure", help="configures petctl")

    # -----------------------------------------
    # Setup
    # -----------------------------------------
    parser_setup = subparser.add_parser(
        "setup", help="creates necessary aws resources and outputs a specs file"
    )
    parser_setup.add_argument(
        "--region", default="us-west-2", help="aws region to setup on"
    )
    parser_setup.add_argument(
        "--s3_bucket",
        help="s3 bucket to use for running petctl (if empty, one is created)",
    )
    parser_setup.add_argument(
        "--efs_id", help="efs id to use, if empty, one is created"
    )

    petctl_args, script_args = split_args(args[1:])
    parsed = parser.parse_args(petctl_args)
    parsed.script_args = script_args
    return parsed


def load_specs_json(file):
    log.info(f"Loading launch specs from: {args.specs_file}")
    with open(file) as f:
        return json.load(f)


def run_job(session, specs_json, args):
    job_name = args.name
    script_args = args.script_args

    rdzv_specs = specs_json["rdzv"]
    worker_specs = specs_json["worker"]

    script_url = urlparse(args.script_path)
    scheme = script_url.scheme
    if scheme == "docker":
        # docker://tmp/script.py -> tmp/script.py (relative to working dir in docker)
        # docker:///tmp/script.py -> /tmp/script.py (absolute path in docker)
        script = script_url.netloc + script_url.path
    elif scheme == "s3":
        # fetch_and_run supports s3:// so just pass through
        script = args.script_path
    else:
        s3_bucket = worker_specs["s3_bucket"]
        s3_prefix = worker_specs["s3_prefix"]
        script = S3(session).cp(args.script_path, s3_bucket, f"{s3_prefix}/{job_name}")

    asg = AutoScalingGroup(session)
    rdzv_asg_name = f"{job_name}_rdzv"
    worker_asg_name = f"{job_name}_worker"

    # create a single node asg to host the etcd server for rendezvous
    etcd_server_hostname = asg.create_asg_sync(rdzv_asg_name, size=1, **rdzv_specs)[0]
    rdzv_endpoint = f"{etcd_server_hostname}:2379"

    # allow overriding instance types from cli
    if args.instance_type:
        worker_specs["instance_type"] = args.instance_type
    worker_specs["rdzv_endpoint"] = rdzv_endpoint
    worker_specs["job_name"] = job_name
    worker_specs["script"] = script
    worker_specs["args"] = " ".join(script_args)
    worker_specs["user"] = getpass.getuser()

    instance_type = worker_specs["instance_type"]
    script_args_str = worker_specs["args"]

    log.info(
        f"\n------------------------------------------------------------------\n"
        f"Starting job...\n"
        f"  job name     : {job_name}\n"
        f"  instance type: {instance_type}\n"
        f"  size         : {args.size} (min={args.min_size}, max={args.max_size})\n"
        f"  rdzv endpoint: {rdzv_endpoint}\n"
        f"  cmd          : {script}\n"
        f"  cmd args     : {script_args_str}\n"
        f"------------------------------------------------------------------\n"
    )

    asg.create_asg(
        worker_asg_name, args.size, args.min_size, args.max_size, **worker_specs
    )


def kill_job(session, specs_json, args):
    job_name = args.job_name
    log.info(f"Killing job {job_name}")
    asg = AutoScalingGroup(session)
    asg.delete_asg(f"{job_name}_rdzv")
    asg.delete_asg(f"{job_name}_worker")


def upload_script(session, specs_json, args):
    script_path = args.script_path
    s3_dest = args.s3_dest

    if not s3_dest:
        s3_bucket = specs_json["s3_bucket"]
        s3_prefix = os.path.join(specs_json["s3_prefix"], getpass.getuser())
    else:
        s3_bucket = urlparse(s3_dest).netloc
        s3_prefix = urlparse(s3_dest).path.strip("/")

    log.info(f"Uploading: {script_path} to s3://{s3_bucket}/{s3_prefix}")
    s3 = S3(session)
    url = s3.cp(script_path, s3_bucket, s3_prefix)
    log.info(f"Finished uploading to: {url}")


def list_hosts(session, specs_json, args):
    job_name = args.job_name
    asg = AutoScalingGroup(session)
    asgs = [f"{job_name}_rdzv", f"{job_name}_worker"]
    hosts = {}

    for asg_name in asgs:
        instance_ids, hostnames = asg.list_hostnames(asg_name)
        hosts[asg_name] = zip(instance_ids, hostnames)

    print(f"\n--------------------------------------------------------------")
    for asg_name in hosts:
        print(f"Hosts in {asg_name}:")
        for i, host in enumerate(hosts[asg_name], start=1):
            instance_id = host[0]
            public_dns = host[1]
            print(f"  {i}) {instance_id} ({public_dns})")
        print(f"--------------------------------------------------------------")
    print("To SSH run:")
    print(f"\taws ssm start-session --target <instance_id>")
    print(f"--------------------------------------------------------------")


def configure(args):
    """
    Configures petctl. Writes a simple json config file indicating
    the specs file to use and the aws region to the petctl config directory
    (default ~/.petctl). Prompts the user to input the specs file location
    and aws region.
    """
    while True:
        specs_file = input(
            "Absolute path to specs file (e.g. /home/${USER}/specs.json): "
        )
        if os.path.isfile(specs_file):
            break
        print(f"[{specs_file}] does not exist! Provide an existing path")

    while True:
        region = input("Default aws region to use (e.g. us-west-2): ")
        if region:
            break
        print("AWS region cannot be empty!")

    write_config_file(region, specs_file)
    log.info(f"Configuration complete. petctl config file: {PETCTL_CONFIG_FILE}")


def setup(args):
    """
    Similar to config but creates AWS resources using cfn template
    and based on the cfn stack output, creates the specs file for the user,
    then writes petctl config.
    """
    region = args.region
    s3_bucket_name = args.s3_bucket
    efs_id = args.efs_id
    os.makedirs(PETCTL_DIR, exist_ok=True)
    session = auth.get_session(region)
    cfn = CloudFormation(session)
    cfn.create_specs_file(SPECS_FILE, s3_bucket_name, efs_id)
    write_config_file(region, SPECS_FILE)
    log.info(f"Setup complete. petctl config file: {PETCTL_CONFIG_FILE}")


def write_config_file(region, specs_file):
    petctl_config = {"specs_file": specs_file, "region": region}
    os.makedirs(PETCTL_DIR, exist_ok=True)
    with open(PETCTL_CONFIG_FILE, "w+") as f:
        json.dump(petctl_config, f, indent=4)


def load_configuration():
    if os.path.isfile(PETCTL_CONFIG_FILE):
        with open(PETCTL_CONFIG_FILE) as f:
            return json.load(f)
    else:
        return {}


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.INFO, format="[%(levelname)s] %(asctime)s %(module)s: %(message)s"
    )

    petctl_configs = load_configuration()
    args = parse_arguments(sys.argv, **petctl_configs)

    if args.command == "setup":
        args = parse_arguments(sys.argv)
        setup(args)
    elif args.command == "configure":
        configure(args)
    else:
        log.info(
            f"{PETCTL_CONFIG_FILE} not found or is empty,"
            f" consider running: petctl setup|configure"
        )
        region = args.region
        specs_json = load_specs_json(args.specs_file)
        session = auth.get_session(region)
        args.func(session, specs_json, args)



#!/usr/bin/env/python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import torch.distributed as dist


class RendezvousParameters(object):

    __slots__ = (
        "rdzv_backend",
        "rdzv_endpoint",
        "run_id",
        "min_workers",
        "max_workers",
        "config",
    )

    def __init__(
        self,
        rdzv_backend: str,
        rdzv_endpoint: str,
        run_id: str,
        min_workers: int,
        max_workers: int,
        config=None,
    ):
        """
        rdzv_backend (str): The rdzv_backend that is used to register
                            rendezvous, e.g. etcd
        rdzv_endpoint (str): The rdzv_endpoint of the rendezvous. Usually it is a string
                        in the format host:port
        run_id (str): The unique job id that is used in rendezvous.
        min_workers (int): The min amount of workers that is required to
                           complete rendezvous.
        max_workers (int): The max amount of workers that can participate in
                           rendezvous.
        config (str): A comma-separated list of additional  configuration
                      of the following format: key1=value1,key2=value2
        """
        self.rdzv_backend = rdzv_backend
        self.rdzv_endpoint = rdzv_endpoint
        self.run_id = run_id
        self.min_workers = min_workers
        self.max_workers = max_workers
        self.config = config if config is not None else ""

        assert 0 < min_workers <= max_workers
        assert rdzv_backend is not None


def _construct_rendezvous_url(params: RendezvousParameters):
    url_suffix = ""
    for kv in params.config.split(","):
        if kv:
            conf_key, conf_val = kv.split("=")
            url_suffix += f"&{conf_key}={conf_val}"

    return (
        f"{params.rdzv_backend}://{params.rdzv_endpoint}/{params.run_id}"
        f"?min_workers={params.min_workers}"
        f"&max_workers={params.max_workers}"
        f"{url_suffix}"
    )


def get_rendezvous(params: RendezvousParameters):
    """
    Given a rdzv parameters, tries to construct a url and find
    the rendezvous handler in using pytorch.distributed.
    The handler should be registerred before this method is invoked.
    The registration of the handlers are done autmatically during
    the module import.
    """
    url = _construct_rendezvous_url(params)
    return dist.rendezvous(url)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.


import getpass
import logging
import os
import random
import string

from jinja2 import Template
from util import wait_for


log = logging.getLogger(__name__)


class CloudFormation:
    def __init__(self, session):
        self._session = session
        self._cfn = session.client("cloudformation")

    def create_specs_file(self, specs_file, s3_bucket_name, efs_id):
        username = getpass.getuser()
        rand = "".join(random.choices(string.ascii_uppercase + string.digits, k=5))
        hash = f"{username}-{rand}"
        stack_name = f"torchelastic-{hash}"
        this_dir = os.path.dirname(__file__)
        cfn_template = os.path.join(this_dir, "cfn/setup.yml")
        sample_specs = os.path.join(this_dir, "config/sample_specs.json")

        params = {
            "WorkerRoleName": f"torchelastic_worker_role-{hash}",
            "RendezvousRoleName": f"torchelastic_rendezvous_role-{hash}",
        }

        if s3_bucket_name:
            params["S3BucketName"] = s3_bucket_name
        if efs_id:
            params["EFSFileSystemId"] = efs_id

        self.create_stack(stack_name, cfn_template, **params)

        for _ in wait_for(
            f"cfn stack: {stack_name} to create", timeout=600, interval=2
        ):
            status, outputs = self.describe_stack(stack_name)
            if status == "CREATE_COMPLETE":
                break
            elif status == "CREATE_FAILED" or status.startswith("ROLLBACK_"):
                # when stack creation fails cfn starts rolling the stack back
                raise RuntimeError(
                    f"Error creating stack {stack_name}, status = {status}"
                )

        outputs["User"] = username

        log.info(f"Writing specs file to: {specs_file}")
        with open(sample_specs) as f:
            specs_template = Template(f.read())
            specs_template.stream(**outputs).dump(specs_file)

    def describe_stack(self, stack_name):
        describe_res = self._cfn.describe_stacks(StackName=stack_name)

        stacks = describe_res["Stacks"]
        if len(stacks) > 1:
            raise RuntimeError(f"Found more than one stack with name {stack_name}")

        stack_desc = stacks[0]
        status = stack_desc["StackStatus"]

        # cfn outputs an array of maps, each element in the array is
        # a single output of the form "{OutputKey: <key>, OutputValue: <value>}"
        # simplify to a map of <key>, <value>  pairs
        outputs = {}
        if "Outputs" in stack_desc:
            for cfn_output in stack_desc["Outputs"]:
                key = cfn_output["OutputKey"]
                value = cfn_output["OutputValue"]
                outputs[key] = value
        return status, outputs

    def create_stack(self, stack_name, cfn_template, **params):
        log.info(f"Creating cloudformation stack with template: {cfn_template}")

        with open(cfn_template) as f:
            template_body = f.read()

        cfn_parameters = []
        for key, value in params.items():
            cfn_parameters.append({"ParameterKey": key, "ParameterValue": value})

        res = self._cfn.create_stack(
            StackName=stack_name,
            TemplateBody=template_body,
            Capabilities=["CAPABILITY_NAMED_IAM"],
            Parameters=cfn_parameters,
        )

        return res["StackId"]



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from .session import AwsSessionProvider


def get_session(region):
    return AwsSessionProvider().get_session(region)


try:
    from .static_init import *  # noqa: F401 F403
except ModuleNotFoundError:
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
import shutil
import tarfile as tar
import tempfile


log = logging.getLogger(__name__)


class S3:
    def __init__(self, session):
        self._session = session
        self._s3 = session.client("s3")

    def cp(self, target_path, bucket, key):
        """
        Uploads target_path to s3://bucket/key. If the target_path is a file
        then uploads to s3://bucket/key/file_name, if the target_path is a
        directory, then a tarball is created with the contents of target_path
        and uploaded to s3://bucket/key/dir_name.tar.gz. The tar is created as
        if created by running the command:

        cd target_path && tar xzf /tmp/$(basename target_path).tar.gz *

        Returns the destination s3 url
        """

        target_basename = os.path.basename(target_path)

        if os.path.isdir(target_path):
            tmpdir = tempfile.mkdtemp(prefix="petctl_")
            tar_basename = f"{target_basename}.tar.gz"
            tar_file = os.path.join(tmpdir, tar_basename)
            log.info(f"Compressing {target_path} into {tar_basename}")
            with tar.open(tar_file, "x:gz") as f:
                f.add(target_path, arcname="", recursive=True)

            dest_key = f"{key}/{tar_basename}"
            target_file = tar_file
        else:
            tmpdir = None
            dest_key = f"{key}/{target_basename}"
            target_file = target_path

        log.info(f"Uploading {target_file} to s3://{bucket}/{dest_key}")
        self._s3.upload_file(target_file, bucket, dest_key)

        if tmpdir:
            log.info(f"Deleting tmp dir: {tmpdir}")
            shutil.rmtree(tmpdir)
        return f"s3://{bucket}/{dest_key}"



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import abc

import boto3


class AwsSessionProvider:
    """
    Provides AWS credentials in the form of boto3 Session.
    This class may be sub-classed to provide custom methods
    of getting aws_access_key_id and aws_secret_access_key.
    Child classes are expected to provide overriding implementations
    of the three `_get_*` methods below.

    When used directly, it follows the default credential
    lookup chain as documented in:
    https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html
    """

    def get_session(self, region=None):
        access_key = self._get_access_key()
        secret_key = self._get_secret_key()
        session_token = self._get_session_token()

        # either both access and secret keys are None
        # or both are not None; just check one to assume
        # the presence of the other
        if access_key is None:
            return boto3.session.Session()
        else:
            return boto3.session.Session(
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_key,
                aws_session_token=session_token,
                region_name=region,
            )

    def _get_access_key(self):
        """
        Returns the aws_access_key_id. Override when sub-classing.
        """
        return None

    def _get_secret_key(self):
        """
        Returns the aws_secret_access_key. Override when sub-classing.
        """
        return None

    def _get_session_token(self):
        """
        Returns the aws_session_token. Override when sub-classing.
        """
        return None
